################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import os
import copy
import yaml
from Tensile.Utilities.ConditionalImports import yamlDumper, yamlLoader
import math

class BenchmarkSplitter(object):

    """
    Benchmark splitter class
    Loads in a benchmark yaml file and splits
    it into several smaller benchmarks limited
    by a number of problem sizes entries.
    """

    @staticmethod
    def __readConfigFile(benchmarkConfigFile):
        with open(benchmarkConfigFile) as f:
            data = yaml.load(f, yamlLoader)
        return data

    # data: a loaded .yaml file
    # returns: a list of yaml files that
    # are differentiated by the benchmark problem
    @staticmethod
    def __splitByProblem(data):
        rv = []
        problemKey = "BenchmarkProblems"
        for i in range(len(data[problemKey])):
            result = {}
            for k in data.keys():
                if k == problemKey:
                    result[k] = [copy.deepcopy(data[k][i])]
                else:
                    result[k] = copy.deepcopy(data[k])
            rv.append(result)
        return rv

    # data: a loaded .yaml file, containing one benchmark problem section
    # returns: a list of yaml files that are differentiated by the
    # benchmark groupings
    @staticmethod
    def __splitByBenchmarkGroup(data):
        rv = []
        problemKey = "BenchmarkProblems"

        assert len(data[problemKey]) == 1, "Config file must have one BenchmarkProblems group"

        benchmarkProblems = data[problemKey][0]

        # Find the index of the problem type group
        problemIdx = -1
        for i in range(len(benchmarkProblems)):
            if("OperationType" in benchmarkProblems[i].keys()):
                problemIdx = i
                break

        assert problemIdx != -1, "Could not find problem type group"

        # Split files on the benchmark group
        for i in range(len(benchmarkProblems)):
            if i != problemIdx:
                result = {}
                for k in data.keys():
                    if k == problemKey:
                        # Take only the problem group and one benchmarkgroup
                        result[k] = [[copy.deepcopy(benchmarkProblems[problemIdx]), \
                                    copy.deepcopy(benchmarkProblems[i])]]
                    else:
                        # copy other sections verbatim
                        result[k] = copy.deepcopy(data[k])
                rv.append(result)
        return rv

    # data: a loaded .yaml file, containing one benchmark problem section
    # and one benchmark group section
    # returns: a list of yaml files that are differentiated by the
    # benchmark sizes
    @staticmethod
    def __splitByBenchmarkSizes(data, numSizes=1):
        rv = []
        problemKey = "BenchmarkProblems"

        assert len(data[problemKey]) == 1, "Config file must have one BenchmarkProblems group"

        benchmarkProblems = data[problemKey][0]

        # Find the index of the problem type group
        # and the benchmark group
        problemIdx = -1
        benchmarkIdx = -1
        for i in range(len(benchmarkProblems)):
            groupKeys = benchmarkProblems[i].keys()
            if("OperationType" in groupKeys):
                problemIdx = i
            elif("BenchmarkFinalParameters" in groupKeys):
                benchmarkIdx = i

        assert len(benchmarkProblems) == 2 \
            and problemIdx != -1 \
            and benchmarkIdx != -1, \
            "Config file must have one ProblemType group and one Benchmark group"

        # Grab the problem sizes from the Benchmark group
        benchmarkGroup = benchmarkProblems[benchmarkIdx]
        assert "ProblemSizes" in benchmarkGroup["BenchmarkFinalParameters"][0] \
                and len(benchmarkGroup["BenchmarkFinalParameters"][0]["ProblemSizes"]), \
                "Benchmark group must have non-empty ProblemSizes"

        problemSizesGroup = benchmarkGroup["BenchmarkFinalParameters"][0]["ProblemSizes"]
        problemSizesCount = len(problemSizesGroup)

        numFiles = math.ceil(problemSizesCount / numSizes)

        # Split files on the benchmark sizes
        for i in range(numFiles):
            result = {}
            for k in data.keys():
                if k == problemKey:

                    # Create a new benchmark group that has all the old keys but split the sizes.
                    newBenchmarkGroup = {}
                    for bk in benchmarkGroup.keys():
                        if bk == "BenchmarkFinalParameters":
                            newBenchmarkGroup[bk] = [ {"ProblemSizes": [] } ]
                            for j in range(i*numSizes, min((i+1)*numSizes, problemSizesCount)):
                                newBenchmarkGroup[bk][0]["ProblemSizes"].append(copy.deepcopy(problemSizesGroup[j]))
                        else:
                            newBenchmarkGroup[bk] = copy.deepcopy(benchmarkGroup[bk])

                    result[k] = [[copy.deepcopy(benchmarkProblems[problemIdx]), copy.deepcopy(newBenchmarkGroup)]]
                else:
                    result[k] = copy.deepcopy(data[k])
            rv.append(result)
        return rv

    # filePath: Name of the file (can be a path)
    # suffix: Add a suffix of _## to fileName by default
    # formatting: How the stuffix string is to be treated.
    # If more indices are needed, you can increase them on the formatting.
    @staticmethod
    def __appendFileNameSuffix(filePath, suffix, separator="_", formatting="{:02}"):
        root, ext = os.path.splitext(filePath)
        suffixString = (separator + formatting).format(suffix)
        return root + suffixString + ext

    @staticmethod
    def splitBenchmarkBySizes(configFile, outputDir, numSizes=1, baseFileName="", separator="_", suffixFormat="{:02}"):

        # Use the configFile as base name if none provided
        if baseFileName == "":
            baseFileName = os.path.basename(configFile)

        # If we want to split by sizes, then we must first split files
        # into separate problems, then split those files into separate
        # benchmark groups, then split those files into sizes
        data = BenchmarkSplitter.__readConfigFile(configFile)
        benchmarksByProblem = BenchmarkSplitter.__splitByProblem(data)
        benchmarksByGroup = []
        benchmarksBySize = []
        for problem in benchmarksByProblem:
            benchmarksByGroup += BenchmarkSplitter.__splitByBenchmarkGroup(problem)
        for group in benchmarksByGroup:
            benchmarksBySize += BenchmarkSplitter.__splitByBenchmarkSizes(group, numSizes)

        # outputDir/basefileName_XX.ext
        outputFileBase = os.path.join(outputDir, baseFileName)
        for i in range(len(benchmarksBySize)):
            outFileName = BenchmarkSplitter.__appendFileNameSuffix(outputFileBase, i, separator, suffixFormat)
            with open(outFileName, "w") as f:
                yaml.dump(benchmarksBySize[i], f, yamlDumper)



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Utils import hash_objs, state


class Property:

    @classmethod
    def FromOriginalState(cls, d):
        return cls(d.get('type'), d.get('index'), d.get('value'))

    def __init__(self, tag=None, index=None, value=None):
        self._tag = tag
        self._index = index
        self._value = value

    @property
    def tag(self):
        return self._tag

    @property
    def index(self):
        return self._index

    @property
    def value(self):
        return self._value

    def state(self):
        rv = {'type': self.tag}
        if self.index is not None: rv['index'] = state(self.index)
        if self.value is not None: rv['value'] = state(self.value)
        return rv

    def __eq__(self, other):
        return self.__class__ == other.__class__ and \
               self.tag   == other.tag   and \
               self.value == other.value and \
               self.index == other.index

    def __hash__(self):
        #return hash(self.tag) ^ hash(self.value) ^ hash(self.index)
        return hash_objs(self.tag, self.value, self.index)


class Predicate(Property):

    @classmethod
    def And(cls, predicates):
        predicates = tuple(predicates)
        if len(predicates) == 0:
            return cls('TruePred')
        if len(predicates) == 1:
            return predicates[0]
        return cls('And', value=predicates)

    @classmethod
    def Or(cls, predicates):
        predicates = tuple(predicates)
        if len(predicates) == 0:
            return cls('TruePred')
        if len(predicates) == 1:
            return predicates[0]
        return cls('Or', value=predicates)

    def __lt__(self, other):
        # Ensure TruePred appears last.
        if other.tag == 'TruePred':
            if self.tag == 'TruePred':
                return False
            return True
        if self.tag == 'TruePred':
            return False

        selfValue = self.value
        otherValue = other.value

        if type(selfValue) is dict:
            selfTup = {v for _, v in selfValue.items()}
            selfValue = selfTup

        if type(otherValue) is dict:
            otherTup = {v for _, v in otherValue.items()}
            otherValue = otherTup

        # If neither is a TruePred then just use the default comparison.
        return (self.tag, self.index, selfValue) < (other.tag, other.index, otherValue)



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import os, subprocess, shutil, random, pytest
from filelock import FileLock
from Tensile import ClientWriter, LibraryIO, Common
from Tensile.SolutionStructs import ProblemType, ProblemSizesMock

# 1. Call TensileCreateLibrary
# 2. Get client instance from ClientExecutable()
# 3. Write some problem types/sizes to ClientParameters with writeClientConfig()
# 4. Run client with small subset of GEMM problems
# $ pytest Tensile/Tests -m integration --capture=tee-sys --builddir /tmp/pytest

# TODO fix TensileCreateLibrary build configs:
#   --library-format=msgpack is currently known to fail
#   --short-file-names is currently known to fail

def downloadLogicFiles(logicDir):
  prefix = "library/src/blas3/Tensile/Logic/asm_full"
  testData = {
    "pre_checkin" : [
      "arcturus/arcturus_Cijk_Ailk_Bjlk_SB.yaml",
      "vega20/vega20_Cijk_Ailk_Bjlk_SB.yaml",
      "vega10/vega10_Cijk_Ailk_Bjlk_SB.yaml",
      "arcturus/arcturus_Cijk_Ailk_Bjlk_BBH.yaml",
      "arcturus/arcturus_Cijk_Ailk_Bjlk_HBH.yaml",
      "vega20/vega20_Cijk_Ailk_Bjlk_HBH.yaml",
      "vega10/vega10_Cijk_Ailk_Bjlk_HBH.yaml",
      "hip/hip_Cijk_Ailk_Bjlk_CB.yaml"
    ],
    "quick" : [
      "arcturus/arcturus_Cijk_Ailk_Bjlk_SB.yaml",
      "vega20/vega20_Cijk_Ailk_Bjlk_SB.yaml",
      "vega10/vega10_Cijk_Ailk_Bjlk_SB.yaml"
    ]
  }

  parentDir = os.path.dirname(logicDir)
  shutil.rmtree(logicDir, ignore_errors=True)

  # basically to query the latest zip release weblink, download it and unzip
  # selected files to destination folder
  cmd = """#!/bin/bash
  set -x
  wget -nc https://api.github.com/repos/ROCmSoftwarePlatform/rocBLAS/releases/latest
  weblink=$(grep -oP '(?<="zipball_url": ")[a-zA-Z:/\.\-0-9]*' latest)
  wget -nc $weblink
  archive=$(basename $weblink)
  rootDir=$(zipinfo -1 $archive | head -n 1)
  """
  for schedule in list(testData.keys()):
    for f in testData[schedule]:
      dir = os.path.join(logicDir, schedule)
      Common.ensurePath(dir)
      cmd += "unzip -j -d %s -x $archive ${rootDir}%s\n"%(dir, os.path.join(prefix,f))

  scriptFile = os.path.join(parentDir,"get_logic.sh")
  with open(scriptFile, "w") as file:
    file.write(cmd)
  os.chmod(scriptFile, 0o777)

  subprocess.run(scriptFile, cwd=parentDir, check=True)

@pytest.fixture(scope="session")
def getLogicFileDir(tmp_path_factory, worker_id):

  if worker_id == "master":
    rootTmpDir = tmp_path_factory.getbasetemp()
    destDir = os.path.join(rootTmpDir, "logic_yaml")
    downloadLogicFiles(destDir)
  else:
    rootTmpDir = tmp_path_factory.getbasetemp().parent
    destDir = os.path.join(rootTmpDir, "logic_yaml")
    lockPath = os.path.join(rootTmpDir, "get_logic.lock")

    with FileLock(lockPath):
      if not os.path.isdir(destDir):
        downloadLogicFiles(destDir)

  return destDir

def isSkippedTest(testYamls, mergeFiles, libraryFormat, shortNames, legacyComponents):
  if testYamls == "pre_checkin":
    # for more extensive tests,
    # we run only on single combination of build params
    if (mergeFiles           == True
        and shortNames       == False
        and libraryFormat    == "yaml"
        and legacyComponents == False):
      return False
    else:
      return True
  elif testYamls == "quick":
    return False
  else:
    assert(False) # should've caught all params already

def str2bool(mergeFiles, shortNames, legacyComponents):
  return (True if mergeFiles=="mergeFiles" else False,
          True if shortNames=="shortNames" else False,
          True if legacyComponents=="legacyComponents" else False)

@pytest.mark.parametrize("testYamls",         ["quick", "pre_checkin"])
@pytest.mark.parametrize("mergeFiles",        ["mergeFiles", "noMergeFiles"])
@pytest.mark.parametrize("libraryFormat",     ["yaml", "msgpack"])
@pytest.mark.parametrize("shortNames",        ["shortNames", "noShortName"])
@pytest.mark.parametrize("legacyComponents",  ["legacyComponents", "noLegacyComponents"])
def test_integration(useGlobalParameters, builddir, getLogicFileDir,
                     testYamls, mergeFiles, libraryFormat, shortNames, legacyComponents):
  mergeFiles, shortNames, legacyComponents = str2bool(mergeFiles, shortNames, legacyComponents)

  if isSkippedTest(testYamls, mergeFiles, libraryFormat, shortNames, legacyComponents):
    pytest.skip("only run pre_checkin test in certain combo")

  logicFileDir = os.path.join(getLogicFileDir, testYamls)
  outputDir    = os.path.join(builddir, "lib")

  with useGlobalParameters(OutputPath=outputDir,
                           WorkingPath=outputDir,
                           MergeFiles=mergeFiles,
                           LibraryFormat=libraryFormat,
                           LegacyComponents=legacyComponents,
                           ShortNames=shortNames,
                           GenerateManifestAndExit=False
                           ):
    Common.ensurePath(outputDir)

    createLibraryScript = ClientWriter.getBuildClientLibraryScript(outputDir, logicFileDir)
    subprocess.run(createLibraryScript, cwd=outputDir, check=True)

    coList = []
    libList = []
    coExt = "co"
    libExt = "yaml" if libraryFormat == "yaml" else "dat"
    with open(os.path.join(outputDir,"library","TensileManifest.txt"), "r") as f:
      lines = f.read().split("\n")
      coList = [line for line in lines if coExt in line]
      libList = [line for line in lines if libExt in line]

    logicFiles = [os.path.join(logicFileDir, f) for f in os.listdir(logicFileDir) \
      if (os.path.isfile(os.path.join(logicFileDir, f)) and os.path.splitext(f)[1]==".yaml")]

    clientParametersPaths = []
    isaStr = "".join([str(e) for e in Common.globalParameters["CurrentISA"]])
    for logicFileName in logicFiles:
      (scheduleName, archName, problemType, _, exactLogic, newLibrary) = \
          LibraryIO.parseLibraryLogicFile(logicFileName)
      problemSizes = ProblemSizesMock(random.sample(exactLogic, min(len(exactLogic), 16))) # sample at most 16 problems
      if isaStr in archName:
        clientParametersPaths.append(ClientWriter.writeClientConfig(
                                      forBenchmark=False,
                                      solutions=None,
                                      problemSizes=problemSizes,
                                      stepName=str(ProblemType(problemType)),
                                      stepBaseDir=outputDir,
                                      newLibrary=newLibrary,
                                      configBase="ClientParameters_%s_%s"%(scheduleName, str(ProblemType(problemType))),
                                      codeObjectFiles=coList,
                                      tileAwareSelection=False,
                                      libraryFile=libList[0]))

    forBenchmark = False
    enableTileSelection = False
    returncode = ClientWriter.runClient(logicFileDir, forBenchmark, enableTileSelection, clientParametersPaths)

    assert(returncode == 0)



################################################################################
#
# Copyright (C) 2021-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import ShiftVectorComponents
from ..AsmUtils import inst, vgpr, sgpr, accvgpr, staticMultiply, vectorStaticDivide, vectorStaticRemainder, log2

class ShiftVectorComponentsVALU(ShiftVectorComponents):
    kernel = {"EnableMatrixInstruction": False}

    """
    Shift Vector Components d0,1
    """
    def __call__(self, writer, kernel, tP):
        kStr = ""

        # glvw
        vw = tP["glvw"]
        numVectors = kernel[tP["tt"]]//vw
        # labels
        svrLabels = []
        sviLabels = []
        for i in range(0, vw):
            r = (i+1) % vw
            label = writer.getLabelNum("ShiftVectorComponents%u_R%u"%(tP["idx"], r) )
            svrLabels.append(label)
            tmpLabels = []
            for v in range(0, numVectors):
                label = writer.getLabelNum("ShiftVectorComponents%u_R%u_V%u"%(tP["idx"], r, v) )
                tmpLabels.append(label)
            sviLabels.append(tmpLabels)

        wg = tP["prevWg"] if writer.prefetchAcrossPersistent else tP["wg"]
        # wgMT value
        tmpSgpr = writer.getTmpSgpr(writer.laneSGPRCount).idx()
        tmpVgpr = writer.vgprPool.checkOut(1,"tmpVgpr")
        wgMT = writer.vgprPool.checkOut(1,"wgMT")
        kStr += inst("v_mov_b32", vgpr(wgMT), sgpr(wg), "")
        kStr += inst("v_mul_i32_i24", vgpr(wgMT), hex(-kernel[tP["mt"]]), vgpr(wgMT), \
                "wg*MT")
        kStr += inst("_v_add_co_u32", vgpr(wgMT), writer.vcc, sgpr("SizesFree+%u"%tP["idx"]), \
                vgpr(wgMT), "wgMT = Size - wg*MT")
        kStr += inst("v_mov_b32", vgpr(tmpVgpr), hex(kernel[tP["mt"]]), "MT")
        kStr += inst("v_min_u32"    , vgpr(wgMT), vgpr(tmpVgpr), vgpr(wgMT), "wgMT = (wgMT < MT) ? wgMT : MT" )
        writer.vgprPool.checkIn(tmpVgpr)

        # qReg
        qReg = writer.vgprPool.checkOut(1,"qReg")
        divisor = kernel["VectorWidth"] # vw
        kStr += vectorStaticDivide(qReg, wgMT, divisor, tmpSgpr)

        # rReg
        rReg = writer.vgprPool.checkOut(1,"rReg")
        divisor = vw
        kStr += vectorStaticRemainder(rReg, wgMT, divisor, tmpSgpr)

        # qReg %/ SG
        eReg = writer.vgprPool.checkOut(1,"eReg")
        divisor = kernel[tP["sg"]]
        kStr += vectorStaticRemainder(eReg, qReg, divisor, tmpSgpr)

        if tP["isA"]:
            # thread = serial % SG0
            thread = writer.vgprPool.checkOut(1,"thread")
            divisor = kernel["SubGroup0"]
            kStr += vectorStaticRemainder(thread, "Serial", divisor, tmpSgpr)
            #kStr += dump(vgpr(thread))
            #kStr += dump(vgpr(thread))
        else:
            # thread = (serial / SG0) % SG1
            sd0 = writer.vgprPool.checkOut(1,"sd0")
            divisor = kernel["SubGroup0"]
            kStr += vectorStaticDivide(sd0, "Serial", divisor, tmpSgpr) # thread = serial / SG0
            divisor = kernel["SubGroup1"]
            thread = writer.vgprPool.checkOut(1,"thread")
            kStr += vectorStaticRemainder(thread, sd0, divisor, tmpSgpr) # thread = (serial / SG0) % SG1
            writer.vgprPool.checkIn(sd0)

        # which glvw vector of thread to shift? wgMT / (SG0*VW) -> (wgMT%VW) / glvw
        # (wgMT/(WG0*VW))*(VW/glvw) + (wgMT%VW) / glvw
        if True:#tP["tensorIdx"] > kernel["VectorWidth"]:
            mvReg = writer.vgprPool.checkOut(1,"mvReg")
            divisor = kernel[tP["sg"]]*kernel["VectorWidth"]
            kStr += vectorStaticDivide(mvReg, wgMT, divisor, tmpSgpr)
            if vw < kernel["VectorWidth"]:
                kStr += inst("v_lshlrev_b32", vgpr(mvReg), hex(log2(kernel["VectorWidth"]//vw)), vgpr(mvReg), "vId *= VW/glvw")
        #kStr += dump(vgpr(mvReg))

        vReg = writer.vgprPool.checkOut(1,"vReg")
        divisor = kernel["VectorWidth"]
        kStr += vectorStaticRemainder(vReg, wgMT, divisor, tmpSgpr)
        vRegD = writer.vgprPool.checkOut(1,"vRegD")
        kStr += inst("v_mov_b32", vgpr(vRegD), vgpr(vReg), "duplicate")
        divisor = vw
        kStr += vectorStaticDivide(vReg, vRegD, divisor, tmpSgpr)
        #kStr += dump(vgpr(vReg))

        if True:#tP["tensorIdx"] > kernel["VectorWidth"]:
            kStr += inst("_v_add_co_u32", vgpr(vReg), writer.vcc, vgpr(mvReg), vgpr(vReg), "vId = 2 components")
            writer.vgprPool.checkIn(mvReg)
            writer.vgprPool.checkIn(vRegD)

        # for each remainder, jump
        for r in range(1, vw):
            kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(rReg), \
                    hex(r), "wgMT%%VW == %u"%r )
            kStr += inst("s_cbranch_vccnz label_%04u"\
                    % svrLabels[(r-1)%vw], \
                    "shift d%u r=%u"%(tP["idx"], r))
            #kStr += inst("s_mov_b32", sgpr(sgprLoc), hex(location), \
            #        "location=%u"%location) location *= 2
            #kStr += inst("v_or_b32", vgpr(vgprPath), sgpr(sgprLoc), \
            #        vgpr(vgprPath), "path+=location")
        kStr += inst("s_branch label_%04u"%svrLabels[vw-1], \
                "no shifting" )

        # code blocks for shifting
        for r in range(1, vw):
            kStr += writer.comment3("shift d%u r=%u"%(tP["idx"], r))
            kStr += "label_%04u:%s" % (svrLabels[r-1], writer.endLine)
            #if r==3 and tP["isA"]:
            #    kStr += dump(vgpr("Serial"))

            # for each vector index, jump
            for vectorIdx in range(0, numVectors):
                kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(vReg), \
                        hex(vectorIdx), "wgMT/(SG*VW) == %u"%vectorIdx )
                kStr += inst("s_cbranch_vccnz label_%04u"\
                        % sviLabels[(r-1)%vw][vectorIdx], \
                        "shift d%u, r=%u, v=%u"%(tP["idx"], r, vectorIdx))

            # code blocks for shifting vector
            for vectorIdx in range(0, numVectors):
                kStr += writer.comment("shift d%u r=%u v=%u"%(tP["idx"], r, vectorIdx))
                kStr += "label_%04u:%s" % (sviLabels[r-1][vectorIdx], writer.endLine)
                # mask if last thread in thread#-tile column
                kStr += inst("_v_cmpx_eq_u32", sgpr(tmpSgpr,writer.laneSGPRCount), vgpr(thread), \
                    vgpr(eReg), "serial % SG == (wgMT/VECTOR_WIDTH)%SG" )
                tto = kernel["ThreadTile%u"%((tP["idx"]+1)%2)] # thread tile orthogonal
                for tt in range(0, tto):
                    for s in range(0, r):
                        if tP["isA"]: # shift d0
                            dst = (s) \
                                + vectorIdx * vw + tt * kernel["ThreadTile0"]
                            src = (s+vw-r) \
                                + vectorIdx * vw + tt * kernel["ThreadTile0"]
                            comment = "rC[%u+%u*VW+%u*TT%s] = rC[%u+%u*VW+%u*TT%s]" \
                                % (s, vectorIdx, tt, writer.tileChar0, \
                                s+vw-r, vectorIdx, tt, writer.tileChar0 )
                        else: # shift d1
                            dst = (tt) \
                                + vectorIdx*vw*kernel["ThreadTile0"] + s * kernel["ThreadTile0"]
                            src = (tt) \
                                + vectorIdx * vw*kernel["ThreadTile0"] + (s+vw-r) * kernel["ThreadTile0"]
                            comment = "rC[%u+%u*TT%s*VW+%u*TT%s] = rC[%u+%u*TT%s*VW+%u*TT%s]" \
                                % (tt, vectorIdx, writer.tileChar0, s, writer.tileChar0, \
                                tt, vectorIdx, writer.tileChar0, \
                                s+vw-r, writer.tileChar0)

                        kStr += "// src=%u, dst=%u\n" % (src,dst)

                        # f16
                        #jgolds I think this should be bpeCinternal
                        if writer.bpeCinternal == 2:
                            srcVgpr = writer.startVgprValuC+src*writer.bpeCinternal//writer.bpr
                            dstVgpr = writer.startVgprValuC+dst*writer.bpeCinternal//writer.bpr
                            kStr += "// %u, %u, %u, %u, %u, %u\n" % (r, vectorIdx, tt, s, dst, src)
                            if tP["isA"]: # f16 d0
                                if r % 2 == 0: # even shift can use mov_b32
                                    if s % 2 == 0:
                                        kStr += inst("v_mov_b32", vgpr(dstVgpr), \
                                                vgpr(srcVgpr), comment)
                                    else:
                                        pass # above command performs two moves
                                else: # odd shift
                                    srcLo = src % 2 == 0 # even
                                    dstLo = dst % 2 == 0 # even
                                    kStr += "// srcLo=%u, dstLo=%u\n" % (srcLo,dstLo)
                                    if dstLo: # hi src to lo dst; can clobber hi bits
                                        kStr += inst("v_lshrrev_b32", vgpr(dstVgpr), \
                                                hex(16), vgpr(srcVgpr), "hi16 -> lo16")
                                    else: # dstHi; cannot clobber lo bits
                                        tmpSrcVgpr = writer.vgprPool.checkOut(1,"tmpSrcVgpr")
                                        # zero out dst hi bits -> dst
                                        kStr += inst("v_and_b32", vgpr(dstVgpr), \
                                                "0x0000FFFF", vgpr(dstVgpr), "zero out dst hi16")
                                        if srcLo: # lo src to hi dst
                                            # left shift src 16 bits -> tmpSrc
                                            kStr += inst("v_lshlrev_b32", vgpr(tmpSrcVgpr), \
                                                    hex(16), vgpr(srcVgpr), "left shift src 16 bits")
                                        else: # hi src to hi dst
                                            # zero out src lo bits -> tmpSrc
                                            kStr += inst("v_and_b32", vgpr(srcVgpr), \
                                                    "0xFFFF0000", vgpr(tmpSrcVgpr), "zero out src lo16")
                                        # dst = tmpSrc | dst
                                        kStr += inst("v_or_b32", vgpr(dstVgpr), \
                                                vgpr(tmpSrcVgpr), vgpr(dstVgpr), "dst = tmpSrc | dst")
                                        writer.vgprPool.checkIn(tmpSrcVgpr)

                            else: # f16 d1
                                if tt%2==0:
                                    kStr += inst("v_mov_b32", vgpr(dstVgpr), \
                                            vgpr(srcVgpr), comment)
                                else:
                                    pass # above shift moves two f16

                        # f32 or larger
                        else:
                            for i in range(0, writer.bpeCinternal//writer.bpr):
                                kStr += inst("v_mov_b32", vgpr(writer.startVgprValuC+dst*writer.bpeCinternal//writer.bpr+i), \
                                        vgpr(writer.startVgprValuC+src*writer.bpeCinternal//writer.bpr+i), comment)

                # end shift reset mask and jump out
                all1mask = "0xFFFFFFFF" if (kernel["WavefrontSize"] == 32) else "0xFFFFFFFFFFFFFFFF"
                kStr += inst("s_mov_b{}".format(kernel["WavefrontSize"]), sgpr(tmpSgpr,writer.laneSGPRCount), \
                        all1mask, "to restore all threads active")
                kStr += inst("s_or_saveexec_b{}".format(kernel["WavefrontSize"]), writer.vcc, sgpr(tmpSgpr,writer.laneSGPRCount), \
                        "all threads active")
                kStr += inst("s_branch label_%04u"%svrLabels[vw-1], \
                        "done shifting" )
        kStr += "label_%04u: // end shift0%s" % (svrLabels[vw-1], writer.endLine)

        # checkin scratch vgprs
        writer.vgprPool.checkIn(wgMT)
        writer.vgprPool.checkIn(qReg)
        writer.vgprPool.checkIn(rReg)
        writer.vgprPool.checkIn(eReg)
        writer.vgprPool.checkIn(thread)
        writer.vgprPool.checkIn(vReg)
        return kStr


class ShiftVectorComponentsMFMA(ShiftVectorComponents):
    kernel = {"EnableMatrixInstruction": True}

    """
    Shift Vector Components d0,1
    """
    def __call__(self, writer, kernel, tP):
        """
           if (glvw > vectorwidth * continuousOutput * threadsInCoal)
               use all thread algorithm
           else
               use partial thread algorithm
        """

        # common parameter
        tc              = tP["tensorChar"]
        glvw            = tP["glvw"]
        numThreadInWave = writer.kernel["WavefrontSize"]
        vectorWidth     = kernel["VectorWidth%s"%tc]

        # use to handle MatrixInst 4x4
        matrixInstM     = kernel["MatrixInstM"] * kernel["MatrixInstBM"] if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstM"]
        matrixInstN     = kernel["MatrixInstN"] * kernel["MatrixInstBN"] if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstN"]

        # unify process for dimension M/N
        matrixInstCoal  = matrixInstM if tP["isA"] else matrixInstN
        matrixInstPrep  = matrixInstN if tP["isA"] else matrixInstM

        # unify process for SourceSwap and non-SourceSwap
        conThInProcDim  = kernel["SourceSwap"] ^ tP["isB"] # continuous threads in processed dimension(Coalesced dimension)
        numThreadInCoal = matrixInstCoal if conThInProcDim else (numThreadInWave // matrixInstPrep)
        numContOutCoal  = vectorWidth if conThInProcDim else kernel["MIOutputVectorWidth"] * vectorWidth
        allContOutCoal  = numContOutCoal

        if (glvw > (allContOutCoal * numThreadInCoal)):
           return self.ShiftVectorComponentsMFMAAllThread(writer, kernel, tP, vectorWidth)
        else:
           return self.ShiftVectorComponentsMFMAPartialThread (writer, kernel, tP, vectorWidth)


    def ShiftVectorComponentsMFMAPartialThread(self, writer, kernel, tP, vectorWidth):
        """ when we enable shift ptr with vectorwidth(2), we shift global read on edge block when size % vectorwidth != 0.
            For example if M size == 3 vector width == 2, we want to do global read for [0-1] and [2-3].
            But 3 is not in memory object, so we shift to do global read [0-1] and [1-2].
            So buffer become [0, 1, 1, 2], assume result in register is same as input [0, 1, 1, 2]
            We need to shift it back to [0, 1, 2].

            In MFMA outputs, We have numContinuousOutput(4) for each thread.
            We have numThreadInWave(64) threads.
            number of thread in N is sames as kernel["MatrixInstN"] (32)
            number of thread in M is numThreadInWave/numOutputThreads1 = 2
            stride of continuous output for each thread (numSubOutputPerWave0) is numOutputThreads0 * numContinuousOutput, (8).
            we have numSubOutputGroupsPerWave0 which is 4 (kernel[tP["mt"]](64) // numSubOutputPerWave0(8))

            So we do shift back by below algorithm.
            1. check if M_size % GlobalLoadVectorWidth != 0, return if == 0
            2. decide which subgroup we need to shift, M_size(3) means 3/8 = group 0
            3. decide which thread we need to shift, we have different groups of thread, (0-31) for first group, (32-63) for second group.
            4. decide which shift block (subTile1) we want to shift. for ex [0-1], [1-2], we want to shift second subtile
        """

        kStr = ""

        # common parameter
        regPerElem      = kernel["MIRegPerOut"]
        glvw            = tP["glvw"]
        numThreadInWave = writer.kernel["WavefrontSize"]
        accImOffset     = writer.AccVgprImagNumOffset(kernel)

        # use to handle MatrixInst 4x4
        matrixInstM     = kernel["MatrixInstM"] * kernel["MatrixInstBM"] if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstM"]
        matrixInstN     = kernel["MatrixInstN"] * kernel["MatrixInstBN"] if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstN"]
        matrixInstBM    = 1 if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstBM"]
        matrixInstBN    = 1 if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstBN"]

        # unify process for dimension M/N
        matrixInstCoal  = matrixInstM              if tP["isA"] else matrixInstN
        matrixInstPrep  = matrixInstN              if tP["isA"] else matrixInstM
        matrixInstBCoal = matrixInstBM             if tP["isA"] else matrixInstBN
        matrixInstBPrep = matrixInstBN             if tP["isA"] else matrixInstBM
        miWaveGroupCoal = kernel["MIWaveGroup"][0] if tP["isA"] else kernel["MIWaveGroup"][1]
        miWGIdStride    = numThreadInWave          if tP["isA"] else (numThreadInWave * kernel["MIWaveGroup"][0])
        miWaveTitleCoal = kernel["MIWaveTile"][0]  if tP["isA"] else kernel["MIWaveTile"][1]
        miWaveTitlePrep = kernel["MIWaveTile"][1]  if tP["isA"] else kernel["MIWaveTile"][0]

        # unify process for SourceSwap and non-SourceSwap
        conThInProcDim  = kernel["SourceSwap"] ^ tP["isB"] # continuous threads in processed dimension(Coalesced dimension)

        threadInterval  = 1 if conThInProcDim else matrixInstPrep
        numThreadInCoal = matrixInstCoal if conThInProcDim else (numThreadInWave // matrixInstPrep)

        numContOutCoal  = vectorWidth if conThInProcDim else kernel["MIOutputVectorWidth"] * vectorWidth
        allContOutCoal  = numContOutCoal

        OutBlocksInMI   = (vectorWidth * matrixInstCoal * matrixInstPrep) // numThreadInWave // numContOutCoal
        OutBlocksInMI   = 1 if conThInProcDim else OutBlocksInMI

        subMBShapeCoal  = (matrixInstCoal * vectorWidth) if conThInProcDim else ((numThreadInWave // matrixInstPrep) * numContOutCoal)
        MBShapeCoal     = subMBShapeCoal * OutBlocksInMI
        MIBShapeCoal    = MBShapeCoal * matrixInstBCoal
        WGShapeCoal     = MIBShapeCoal * miWaveGroupCoal
        miOuterTTCoal   = miWaveTitleCoal // vectorWidth

        numOutputsPrep  = (matrixInstCoal * matrixInstPrep // numThreadInWave) if conThInProcDim else 1
        numOutputsPrep  = numOutputsPrep * matrixInstBPrep * miWaveTitlePrep
        complexMultiplier = 2 if kernel["ProblemType"]["DataType"].isComplex() else 1

        # unify process for dimension M/N
        regStrideCoal = 1                                                                if tP["isA"] else numOutputsPrep
        regStridePrep = miOuterTTCoal * matrixInstBCoal * OutBlocksInMI * allContOutCoal if tP["isA"] else 1


        # labels for shiftptr
        glvwLabels = []
        MBblockLabels = []
        VWBlockLabels = []
        for i in range(0, glvw): # grvw block
            r = (i+1) % glvw    # r = [1,2,3,...,glvw-1, 0], the last one glvwLabels[glvw-1] stores for r=0 -> no shift
            label = writer.getLabelNum("ShiftVectorComponents%u_GLVW%u" % (tP["idx"], r) )
            glvwLabels.append(label)
            subMBLabels = []
            subVWBlockLabels = []
            for mb in range(0, OutBlocksInMI * matrixInstBCoal * miOuterTTCoal): # unit block of each thread
                label = writer.getLabelNum("ShiftVectorComponents%u_GLVW%u_BM%u" % (tP["idx"], r, mb))
                subMBLabels.append(label)
                sub2VWBlockLabels = []
                for vw in range(0, max(1, allContOutCoal//glvw)): # vw block of glvw
                    label = writer.getLabelNum("ShiftVectorComponents%u_GLVW%u_BM%u_VW%u" % (tP["idx"], r, mb, vw))
                    sub2VWBlockLabels.append(label)
                subVWBlockLabels.append(sub2VWBlockLabels)
            MBblockLabels.append(subMBLabels)
            VWBlockLabels.append(subVWBlockLabels)

        # wgMT value
        tmpSgpr = writer.getTmpSgpr(writer.laneSGPRCount).idx()
        tmpVgpr = writer.vgprPool.checkOut(1)
        wgMT    = writer.vgprPool.checkOut(1)
        wg      = tP["prevWg"] if writer.prefetchAcrossPersistent else tP["wg"]

        # get M size of edge block
        mtReg = writer.vgprPool.checkOut(1)
        kStr += inst("v_mov_b32"    , vgpr(wgMT), sgpr(wg), "")
        kStr += inst("v_mul_i32_i24", vgpr(wgMT), hex(-kernel[tP["mt"]]), vgpr(wgMT), "wg*MT")
        kStr += inst("_v_add_co_u32", vgpr(wgMT), writer.vcc, sgpr("SizesFree+%u"%tP["idx"]), vgpr(wgMT), "wgMT = Size - wg*MT")
        kStr += inst("v_mov_b32"    , vgpr(mtReg), hex(kernel[tP["mt"]]), "MT")
        kStr += inst("v_min_u32"    , vgpr(wgMT), vgpr(mtReg), vgpr(wgMT), "wgMT = (wgMT < MT) ? wgMT : MT" )

        # identify which wave have to process
        wReg = writer.vgprPool.checkOut(1)
        sReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticDivide(tmpVgpr, "Serial", miWGIdStride, tmpSgpr)
        kStr += vectorStaticRemainder(wReg, tmpVgpr, miWaveGroupCoal, tmpSgpr)
        kStr += vectorStaticDivide(tmpVgpr, wgMT, MIBShapeCoal, tmpSgpr)
        kStr += vectorStaticRemainder(sReg, tmpVgpr, miWaveGroupCoal, tmpSgpr)
        kStr += inst("v_cmp_eq_u32" , sgpr(tmpSgpr,writer.laneSGPRCount), vgpr(sReg), vgpr(wReg), "wave_id == block_belong_to_wave?" )
        kStr += inst("v_cndmask_b32", vgpr(wgMT), vgpr(mtReg), vgpr(wgMT), sgpr(tmpSgpr,writer.laneSGPRCount), "wgMT = (wgMT < MT) ? wgMT : MT" )
        writer.vgprPool.checkIn(mtReg)
        writer.vgprPool.checkIn(sReg)

        # mbReg: which mb block meed to shift, mb(matrixInstM*VectorWidth)
        kStr += writer.comment("mbReg: which mb block need to shift, mb(matrixInstCoal(%u) * VectorWidth(%u))" % (matrixInstCoal, vectorWidth))
        mbReg = writer.vgprPool.checkOut(1)
        tReg  = writer.vgprPool.checkOut(1)
        kStr += vectorStaticDivide(mbReg, wgMT, subMBShapeCoal, tmpSgpr)
        kStr += staticMultiply(vgpr(tReg), vgpr(wReg), (matrixInstBCoal * OutBlocksInMI), sgpr(tmpSgpr))
        kStr += inst("_v_sub_u32", vgpr(mbReg), vgpr(mbReg), vgpr(tReg), "")
        writer.vgprPool.checkIn(tReg)

        # gbReg: glvw block id
        kStr += writer.comment("gbReg: glvw block id")
        gbReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticDivide(gbReg, wgMT, glvw, tmpSgpr)

        # tgbReg: thread in glvw block
        kStr += writer.comment("tgbReg: glvw block id")
        tgbReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticDivide(tmpVgpr, "Serial", threadInterval, tmpSgpr)
        kStr += vectorStaticRemainder(tgbReg, tmpVgpr, numThreadInCoal, tmpSgpr)
        kStr += staticMultiply(vgpr(tgbReg), vgpr(tgbReg), allContOutCoal, sgpr(tmpSgpr))
        kStr += vectorStaticDivide(tgbReg, tgbReg, glvw, tmpSgpr)
        kStr += staticMultiply(vgpr(wReg), vgpr(wReg), MIBShapeCoal//glvw, sgpr(tmpSgpr))
        kStr += inst("_v_add_co_u32", vgpr(tgbReg), writer.vcc, vgpr(wReg), vgpr(tgbReg), "tgbReg = (tid_coal * continOut) / GLVW")
        kStr += inst("_v_sub_u32", vgpr(gbReg), vgpr(gbReg), vgpr(tgbReg), "")
        writer.vgprPool.checkIn(wReg)
        writer.vgprPool.checkIn(tgbReg)

        # vw block of glvw
        kStr += writer.comment("vwReg: glvw in which vw block?")
        vwReg = writer.vgprPool.checkOut(1)
        kStr += inst("v_and_b32", vgpr(vwReg), allContOutCoal-1, vgpr(wgMT), "permute register between threads")
        kStr += inst("v_lshrrev_b32", vgpr(vwReg), log2(glvw), vgpr(vwReg), "permute register between threads")

        # rReg : reminder of M_size % vectorwidth
        # decide to jump to block which handle this case, M_size % vector width
        kStr += writer.comment("rReg : reminder of M_size % GlobalLoadVectorWidth")
        rReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticRemainder(rReg, wgMT, glvw, tmpSgpr)
        for r in range(1, glvw):
            kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(rReg), hex(r), "wgMT%%VW == %u"%r )
            kStr += inst("s_cbranch_vccnz label_%04u" % glvwLabels[(r-1)], "branch to shift d%u r=%u"%(tP["idx"], r))
        kStr += inst("s_branch label_%04u"%glvwLabels[glvw-1], "no shifting" )
        writer.vgprPool.checkIn(rReg)

        _, arch2acc = writer.AccToArchMapper(kernel)

        # blocks for handle M_size % vector width
        for r in range(1, glvw):
            kStr += writer.comment3("shift d%u r=%u"%(tP["idx"], r))
            kStr += "label_%04u:%s" % (glvwLabels[r-1], writer.endLine)
            for tt in range(0, miOuterTTCoal):
                for bm in range(0, matrixInstBCoal):
                    for ob in range(0, OutBlocksInMI):
                        label  = ob + OutBlocksInMI * (bm + matrixInstBCoal * tt)
                        target = ob + OutBlocksInMI * (bm + matrixInstBCoal * miWaveGroupCoal * tt)
                        kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(mbReg), hex(target), "")
                        kStr += inst("s_cbranch_vccnz label_%04u" % MBblockLabels[r-1][label], "branch to shift d%u r%u mb%u" % (tP["idx"], r, label))

        for r in range(1, glvw):
            for mb in range(0, miOuterTTCoal * matrixInstBCoal * OutBlocksInMI):
                kStr += writer.comment3("shift d%u r=%u mb=%u"%(tP["idx"], r, mb))
                kStr += "label_%04u: // r%u mb%u %s" % (MBblockLabels[r-1][mb], r, mb, writer.endLine)
                for vw in range(0, max(1, allContOutCoal//glvw)):
                    kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(vwReg), hex(vw), "")
                    kStr += inst("s_cbranch_vccnz label_%04u" % VWBlockLabels[r-1][mb][vw], "branch to shift d%u r%u mb%u vw%u" % (tP["idx"], r, mb, vw))

        # blocks for handle M_size % vector width
        # no need to allocate tReg for the following scenarios
        #   - allContOutCoal==1 and kernel["MIArchVgpr"]
        #   - glvw <= allContOutCoal and glvw<=2 and kernel["MIArchVgpr"]
        needTReg = not (allContOutCoal==1 and kernel["MIArchVgpr"] or \
                        glvw <= allContOutCoal and glvw<=2 and kernel["MIArchVgpr"])
        tReg = None
        if needTReg:
          tReg  = writer.vgprPool.checkOut(min(glvw, allContOutCoal))
        for r in range(1, glvw):
            for tt in range(0, miOuterTTCoal):
                for bm in range(0, matrixInstBCoal):
                    for ob in range(0, OutBlocksInMI):
                        mb = ob + OutBlocksInMI * (bm + matrixInstBCoal * tt)
                        for vw in range(0, max(1, allContOutCoal//glvw)):
                            kStr += writer.comment3("shift d%u r=%u mb=%u vw%d"%(tP["idx"], r, mb, vw))
                            kStr += "label_%04u: // r%u mb%u vw%u %s" % (VWBlockLabels[r-1][mb][vw], r, mb, vw, writer.endLine)
                            kStr += inst("s_mov_b32", sgpr(tmpSgpr), (((ob*subMBShapeCoal + bm*MBShapeCoal + tt*WGShapeCoal) // glvw) + vw), "")
                            kStr += inst("_v_cmpx_eq_u32", sgpr(tmpSgpr, writer.laneSGPRCount), vgpr(gbReg), sgpr(tmpSgpr), "is thread in edge glvw region" )
                            kStr += inst("v_and_b32", vgpr(tmpVgpr), kernel["WavefrontSize"]-1, vgpr("Serial"), "permute register between threads")
                            kStr += inst("v_lshlrev_b32", vgpr(tmpVgpr), log2(writer.bpr), vgpr(tmpVgpr), "permute register between threads")

                            for ot in range(numOutputsPrep):
                                for c  in range(complexMultiplier):
                                    for nr in range(regPerElem):
                                        vgprOffsetForSCIU = 0
                                        if kernel["StoreCInUnroll"] and writer.enableSingleNLLOpt:
                                          # single NLL opt case, use second acc register set
                                          vgprOffsetForSCIU += writer.startaccValuC1
                                        if allContOutCoal==1 and kernel["MIArchVgpr"]:
                                          # if allContOutCoal==1, src and dest are always same
                                          # MIArchVgpr case, move is unnecessary and we can directly update srcVgpr with ds_bpermute_b32
                                          srcVgpr = ((vw * glvw) + allContOutCoal * mb) * regStrideCoal
                                          srcVgpr = srcVgpr + ot * regStridePrep
                                          srcVgpr = arch2acc[srcVgpr] * regPerElem + nr + c * accImOffset + vgprOffsetForSCIU
                                          # allContOutCoal==1 case, crossThread is always non 0 (= glvw-r)
                                          # ds_bpermute_b32 is always necessary
                                          crossThread = (glvw-r) // allContOutCoal
                                          kStr += inst("ds_bpermute_b32", vgpr(srcVgpr), vgpr(tmpVgpr), vgpr(srcVgpr), "offset:{}".format(crossThread*threadInterval*4), "permute edge values")
                                          kStr += inst("s_waitcnt", "0", "wait for swizzle operation")
                                        elif glvw <= allContOutCoal and glvw<=2 and kernel["MIArchVgpr"]:
                                          # if glvw <= allContOutCoal,
                                          #   -> r < glvw <= allContOutCoal, then, min(r, allContOutCoal) = r
                                          #   -> (e+(glvw-r)) <= r-1+(glvw-r) = glvw-1 < allContOutCoal (because glvw<=allContOutCoal)
                                          #   -> crossThread = (e+(glvw-r)) // allContOutCoal = 0 and we do not need ds_bpermute_b32
                                          # MIArchVgpr case, we do not need to move src to tReg.
                                          # Instead, we can directly move src to dst and reduce the amount of v_mov
                                          # To safely reduce v_mov, the range of e should be range(1). Otherwise, mov to dst might overwrite unused src regs
                                          #   -> glvw<=2 ensures the range of e to be range(1)
                                          copyInstStr = "v_mov_b32"
                                          for e in range(min(r, allContOutCoal)):
                                              src = (e+(glvw-r)) % allContOutCoal
                                              srcVgpr = (src + (vw * glvw) + allContOutCoal * mb) * regStrideCoal
                                              srcVgpr = srcVgpr + ot * regStridePrep
                                              srcVgpr = arch2acc[srcVgpr] * regPerElem + nr + c * accImOffset + vgprOffsetForSCIU
                                              srcStr = vgpr(srcVgpr)
                                              dstVgpr = (e + (vw * glvw) + allContOutCoal * mb) * regStrideCoal
                                              dstVgpr = dstVgpr + ot * regStridePrep
                                              dstVgpr = arch2acc[dstVgpr] * regPerElem + nr + c * accImOffset + vgprOffsetForSCIU
                                              dstStr = vgpr(dstVgpr)
                                              kStr += inst(copyInstStr, dstStr, srcStr, "glvw %u mb %u tt1 %u r %u" % (r, mb, ot, nr))
                                        else:
                                          copyInstStr = "v_accvgpr_read_b32" if not kernel["MIArchVgpr"] else "v_mov_b32"
                                          for e in range(min(r, allContOutCoal)):
                                              src = (e+(glvw-r)) % allContOutCoal
                                              srcVgpr = (src + (vw * glvw) + allContOutCoal * mb) * regStrideCoal
                                              srcVgpr = srcVgpr + ot * regStridePrep
                                              srcVgpr = arch2acc[srcVgpr] * regPerElem + nr + c * accImOffset + vgprOffsetForSCIU
                                              srcStr = accvgpr(srcVgpr) if not kernel["MIArchVgpr"] else vgpr(srcVgpr)
                                              kStr += inst(copyInstStr, vgpr(tReg+e), srcStr, "glvw %u mb %u tt1 %u r %u" % (r, mb, ot, nr))

                                          if not kernel["MIArchVgpr"]:
                                              kStr += inst("s_nop", "1", "v_accvgpr read vgpr after write vgpr: 2 wait states")

                                          needWait = False
                                          for e in range(min(r, allContOutCoal)):
                                              crossThread = (e+(glvw-r)) // allContOutCoal
                                              if crossThread != 0:
                                                  kStr += inst("ds_bpermute_b32", vgpr(tReg+e), vgpr(tmpVgpr), vgpr(tReg+e), "offset:{}".format(crossThread*threadInterval*4), "permute edge values")
                                                  needWait = True

                                          if needWait:
                                              kStr += inst("s_waitcnt", "0", "wait for swizzle operation")

                                          copyInstStr = "v_accvgpr_write_b32" if not kernel["MIArchVgpr"] else "v_mov_b32"
                                          for e in range(min(r, allContOutCoal)):
                                              dstVgpr = (e + (vw * glvw) + allContOutCoal * mb) * regStrideCoal
                                              dstVgpr = dstVgpr + ot * regStridePrep
                                              dstVgpr = arch2acc[dstVgpr] * regPerElem + nr + c * accImOffset + vgprOffsetForSCIU
                                              dstStr = accvgpr(dstVgpr) if not kernel["MIArchVgpr"] else vgpr(dstVgpr)
                                              kStr += inst(copyInstStr, dstStr, vgpr(tReg+e), "")

                            # end shift reset mask and jump out
                            all1mask = "0xFFFFFFFF" if (kernel["WavefrontSize"] == 32) else "0xFFFFFFFFFFFFFFFF"
                            kStr += inst("s_mov_b{}".format(kernel["WavefrontSize"]), sgpr(tmpSgpr, writer.laneSGPRCount), all1mask, "to restore all threads active")
                            kStr += inst("s_or_saveexec_b{}".format(kernel["WavefrontSize"]), writer.vcc, sgpr(tmpSgpr,writer.laneSGPRCount), "all threads active")
                            kStr += inst("s_branch label_%04u" % glvwLabels[glvw-1], "done shifting" )
                            kStr += writer.endLine

        kStr += "label_%04u: // end shift0%s" % (glvwLabels[glvw-1], writer.endLine)
        if tReg!= None:
          writer.vgprPool.checkIn(tReg)

        # checkin scratch vgprs
        writer.vgprPool.checkIn(tmpVgpr)
        writer.vgprPool.checkIn(wgMT)
        writer.vgprPool.checkIn(gbReg)
        writer.vgprPool.checkIn(vwReg)
        writer.vgprPool.checkIn(mbReg)

        return kStr


    def ShiftVectorComponentsMFMAAllThread(self, writer, kernel, tP, vectorWidth):
        """
        """

        kStr = ""

        # common parameter
        glvw            = tP["glvw"]
        numThreadInWave = writer.kernel["WavefrontSize"]

        # use to handle MatrixInst 4x4
        matrixInstM     = kernel["MatrixInstM"] * kernel["MatrixInstBM"] if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstM"]
        matrixInstN     = kernel["MatrixInstN"] * kernel["MatrixInstBN"] if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstN"]
        matrixInstBM    = 1 if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstBM"]
        matrixInstBN    = 1 if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstBN"]

        # unify process for dimension M/N
        matrixInstCoal  = matrixInstM              if tP["isA"] else matrixInstN
        matrixInstPrep  = matrixInstN              if tP["isA"] else matrixInstM
        matrixInstBCoal = matrixInstBM             if tP["isA"] else matrixInstBN
        matrixInstBPrep = matrixInstBN             if tP["isA"] else matrixInstBM
        miWaveGroupCoal = kernel["MIWaveGroup"][0] if tP["isA"] else kernel["MIWaveGroup"][1]
        miWGIdStride    = numThreadInWave          if tP["isA"] else (numThreadInWave * kernel["MIWaveGroup"][0])
        miWaveTitleCoal = kernel["MIWaveTile"][0]  if tP["isA"] else kernel["MIWaveTile"][1]
        miWaveTitlePrep = kernel["MIWaveTile"][1]  if tP["isA"] else kernel["MIWaveTile"][0]

        # unify process for SourceSwap and non-SourceSwap
        conThInProcDim  = kernel["SourceSwap"] ^ tP["isB"] # continuous threads in processed dimension(Coalesced dimension)

        threadInterval  = 1 if conThInProcDim else matrixInstPrep
        numThreadInCoal = matrixInstCoal if conThInProcDim else (numThreadInWave // matrixInstPrep)

        numContOutCoal  = vectorWidth if conThInProcDim else kernel["MIOutputVectorWidth"] * vectorWidth
        allContOutCoal  = numContOutCoal

        OutBlocksInMI   = (vectorWidth * matrixInstCoal * matrixInstPrep) // numThreadInWave // numContOutCoal
        OutBlocksInMI   = 1 if conThInProcDim else OutBlocksInMI

        subMBShapeCoal  = (matrixInstCoal * vectorWidth) if conThInProcDim else ((numThreadInWave // matrixInstPrep) * numContOutCoal)
        MBShapeCoal     = subMBShapeCoal * OutBlocksInMI
        MIBShapeCoal    = MBShapeCoal * matrixInstBCoal
        miOuterTTCoal   = miWaveTitleCoal // vectorWidth

        numOutputsPrep  = (matrixInstCoal * matrixInstPrep // numThreadInWave) if conThInProcDim else 1
        numOutputsPrep  = numOutputsPrep * matrixInstBPrep * miWaveTitlePrep

        # unify process for dimension M/N
        regStrideCoal = 1                                                                if tP["isA"] else numOutputsPrep
        regStridePrep = miOuterTTCoal * matrixInstBCoal * OutBlocksInMI * allContOutCoal if tP["isA"] else 1

        # labels for shiftptr
        shiftLabels = []
        GLVWBLKLabels = []
        for i in range(0, glvw): # grvw block
            r = (i+1) % glvw    # r = [1,2,3,...,glvw-1, 0], the last one shiftLabels[glvw-1] stores for r=0 -> no shift
            label = writer.getLabelNum("ShiftVectorComponents%u_shift%u" % (tP["idx"], r) )
            shiftLabels.append(label)
            subGLVWBLKLabels = []
            for glvwBlk in range(0, (MIBShapeCoal // glvw) * miOuterTTCoal):
                label = writer.getLabelNum("ShiftVectorComponents%u_shift%u_glvwblk%u" % (tP["idx"], r, glvwBlk))
                subGLVWBLKLabels.append(label)
            GLVWBLKLabels.append(subGLVWBLKLabels)

        # wgMT value
        wg      = tP["prevWg"] if writer.prefetchAcrossPersistent else tP["wg"]
        tmpSgpr = writer.getTmpSgpr(writer.laneSGPRCount).idx()

        tmpVgpr = writer.vgprPool.checkOut(1)
        wgMT    = writer.vgprPool.checkOut(1)
        mtReg   = writer.vgprPool.checkOut(1)

        # get M size of edge block
        kStr += writer.comment1("check which macro tile need to shift")
        kStr += inst("v_mov_b32"    , vgpr(wgMT), sgpr(wg), "")
        kStr += inst("v_mul_i32_i24", vgpr(wgMT), hex(-kernel[tP["mt"]]), vgpr(wgMT), "wg*MT")
        kStr += inst("_v_add_co_u32", vgpr(wgMT), writer.vcc, sgpr("SizesFree+%u"%tP["idx"]), vgpr(wgMT), "wgMT = Size - wg*MT")
        kStr += inst("v_mov_b32"    , vgpr(mtReg), hex(kernel[tP["mt"]]), "MT")
        kStr += inst("v_min_u32"    , vgpr(wgMT), vgpr(mtReg), vgpr(wgMT), "wgMT = (wgMT < MT) ? wgMT : MT" )
        kStr += writer.endLine

        # identify which wave have to process
        kStr += writer.comment1("check which wave need to shift")
        wReg = writer.vgprPool.checkOut(1)
        sReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticDivide(tmpVgpr, "Serial", miWGIdStride, tmpSgpr)
        kStr += vectorStaticRemainder(wReg, tmpVgpr, miWaveGroupCoal, tmpSgpr)
        kStr += vectorStaticDivide(tmpVgpr, wgMT, MIBShapeCoal, tmpSgpr)
        kStr += vectorStaticRemainder(sReg, tmpVgpr, miWaveGroupCoal, tmpSgpr)
        kStr += inst("v_cmp_eq_u32" , sgpr(tmpSgpr,writer.laneSGPRCount), vgpr(sReg), vgpr(wReg), "wave_id == block_belong_to_wave?" )
        kStr += inst("v_cndmask_b32", vgpr(wgMT), vgpr(mtReg), vgpr(wgMT), sgpr(tmpSgpr,writer.laneSGPRCount), "wgMT = (wave_id == block_belong_to_wave) ? wgMT : MT" )
        kStr += writer.endLine

        # glveblkid
        kStr += writer.comment1("get id of which glvw block need to shift")
        glvwblkidReg = writer.vgprPool.checkOut(1)
        kStr += inst("v_mul_i32_i24", vgpr(glvwblkidReg), hex(-MIBShapeCoal), vgpr(wReg), "wg * MIB")
        kStr += inst("_v_add_co_u32", vgpr(glvwblkidReg), writer.vcc, vgpr(glvwblkidReg), vgpr(wgMT), "wgMT = Size - wg*MIB")
        kStr += vectorStaticDivide(glvwblkidReg, glvwblkidReg, glvw, tmpSgpr, "glvw block id")
        kStr += writer.endLine

        # rReg : reminder of M_size % vectorwidth
        # decide to jump to block which handle this case, M_size % vector width
        kStr += writer.comment1("dispatch to different shift block for shift")
        rReg = writer.vgprPool.checkOut(1)
        kStr += vectorStaticRemainder(rReg, wgMT, glvw, tmpSgpr)
        for r in range(1, glvw):
            kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(rReg), hex(r), "wgMT%%GLVW == %u"%r )
            kStr += inst("s_cbranch_vccnz label_%04u" % shiftLabels[(r-1)], "branch to shift d%u r=%u"%(tP["idx"], r))
        kStr += inst("s_branch label_%04u" % shiftLabels[glvw-1], "no shifting" )
        writer.vgprPool.checkIn(rReg)

        # blocks for handle M_size % vector width
        glvwBlkInMIB = MIBShapeCoal // glvw
        numRegInGlvwblkCoal = glvw // numThreadInCoal
        numRegInMIBCoal = MIBShapeCoal // numThreadInCoal
        for r in range(1, glvw):
            kStr += writer.comment3("shift d%u shift=%u"%(tP["idx"], r))
            kStr += "label_%04u:%s" % (shiftLabels[r-1], writer.endLine)
            for tt in range(0, miOuterTTCoal):
                for glvwBlk in range(0, glvwBlkInMIB):
                    label  = glvwBlk + tt * glvwBlkInMIB
                    target = glvwBlk + tt * glvwBlkInMIB * miWaveGroupCoal
                    kStr += inst("v_cmp_eq_u32", writer.vcc, vgpr(glvwblkidReg), hex(target), "")
                    kStr += inst("s_cbranch_vccnz label_%04u" % GLVWBLKLabels[r-1][label], "branch to shift d%u shift%u glvwblk%u" % (tP["idx"], r, target))

        _, arch2acc = writer.AccToArchMapper(kernel)

        permuteIndexReg = writer.vgprPool.checkOut(1)
        threadIdInCoalReg = writer.vgprPool.checkOut(1)
        movReg = writer.vgprPool.checkOut(numContOutCoal*numOutputsPrep)
        kStr += writer.comment3("Tony Reg %d-%d"%(movReg, movReg+15))
        for shift in range(1, glvw):
            for tt in range(0, miOuterTTCoal):
                for glvwBlk in range(0, glvwBlkInMIB):
                    label = glvwBlk + tt * glvwBlkInMIB
                    kStr += writer.comment3("shift d%u shift=%u glvwblk=%u"%(tP["idx"], shift, glvwBlk))
                    kStr += "label_%04u:%s" % (GLVWBLKLabels[shift-1][label], writer.endLine)
                    kStr += inst("v_and_b32", vgpr(permuteIndexReg), kernel["WavefrontSize"]-1, vgpr("Serial"), "permute register between threads")
                    kStr += staticMultiply(vgpr(permuteIndexReg), vgpr(permuteIndexReg), writer.bpr, sgpr(tmpSgpr), "permute register between threads")

                    kStr += vectorStaticDivide(tmpVgpr, "Serial", threadInterval, tmpSgpr)
                    kStr += vectorStaticRemainder(threadIdInCoalReg, tmpVgpr, numThreadInCoal, tmpSgpr)

                    for dstMbblkId in range(glvw//(numContOutCoal*numThreadInCoal)):
                        for dstThreadId in range(numThreadInCoal):
                            skip = True
                            copyInstStr = "v_accvgpr_read_b32" if not kernel["MIArchVgpr"] else "v_mov_b32"
                            for dstContId in range(numContOutCoal):
                                dst = dstContId + dstThreadId * numContOutCoal + dstMbblkId * numThreadInCoal * numContOutCoal
                                src = dst + (glvw - shift)
                                if (src < glvw):
                                    skip = False
                                    srcContId   = src % numContOutCoal
                                    srcThreadId = (src // numContOutCoal) % numThreadInCoal
                                    srcMbblkId  = src // (numContOutCoal * numThreadInCoal)
                                    for ot in range(numOutputsPrep):
                                      movRegId    = movReg + dstContId + ot * numContOutCoal
                                      srcGpr      = srcContId + srcMbblkId * numContOutCoal + glvwBlk * numRegInGlvwblkCoal + tt * numRegInMIBCoal
                                      srcGpr      = srcGpr * regStrideCoal + ot * regStridePrep
                                      srcGpr      = arch2acc[srcGpr]
                                      srcGprStr   = accvgpr(srcGpr) if not kernel["MIArchVgpr"] else vgpr(srcGpr)
                                      kStr += inst(copyInstStr, vgpr(movRegId), srcGprStr, "")

                            if not skip:
                                kStr += inst("s_nop", "1", "v_accvgpr read vgpr after write vgpr: 2 wait states")

                            needWait = False
                            for dstContId in range(numContOutCoal):
                                dst = dstContId + dstThreadId * numContOutCoal + dstMbblkId * numThreadInCoal * numContOutCoal
                                src = dst + (glvw - shift)
                                if (src < glvw):
                                    srcContId   = src % numContOutCoal
                                    srcThreadId = (src // numContOutCoal) % numThreadInCoal
                                    srcMbblkId  = src // (numContOutCoal * numThreadInCoal)
                                    srcGpr      = srcContId + srcMbblkId * numContOutCoal
                                    if dstThreadId != srcThreadId:
                                        needWait = True
                                        permuteOffset = (((srcThreadId - dstThreadId) * threadInterval) % kernel["WavefrontSize"]) * 4
                                        for ot in range(numOutputsPrep):
                                            movRegId    = movReg + dstContId + ot * numContOutCoal
                                            kStr += inst("ds_bpermute_b32", vgpr(movRegId), vgpr(permuteIndexReg), vgpr(movRegId), f"offset:{permuteOffset}", "permute edge values")

                            if needWait:
                                kStr += inst("s_waitcnt", "lgkmcnt(0)", "wait for swizzle operation")

                            if not skip:
                                kStr += inst("s_mov_b32", sgpr(tmpSgpr), dstThreadId, "which thread need to shfit in this block")
                                kStr += inst("_v_cmpx_eq_u32", sgpr(tmpSgpr, writer.laneSGPRCount), vgpr(threadIdInCoalReg), sgpr(tmpSgpr), "is thread in edge glvw region" )
                                kStr += inst("s_nop", "3", "wait for exec mask")

                            copyInstStr = "v_accvgpr_write_b32" if not kernel["MIArchVgpr"] else "v_mov_b32"
                            for dstContId in range(numContOutCoal):
                                dst = dstContId + dstThreadId * numContOutCoal + dstMbblkId * numThreadInCoal * numContOutCoal
                                src = dst + (glvw - shift)
                                if (src < glvw):
                                    for ot in range(numOutputsPrep):
                                        movRegId    = movReg + dstContId + ot * numContOutCoal
                                        dstGpr      = dstContId + dstMbblkId * numContOutCoal + glvwBlk * numRegInGlvwblkCoal + tt * numRegInMIBCoal
                                        dstGpr      = dstGpr * regStrideCoal + ot * regStridePrep
                                        dstGpr      = arch2acc[dstGpr]
                                        dstGprStr   = accvgpr(dstGpr) if not kernel["MIArchVgpr"] else vgpr(dstGpr)
                                        kStr += inst(copyInstStr, dstGprStr, vgpr(movRegId), "")

                            if not skip:
                                all1mask = "0xFFFFFFFF" if (kernel["WavefrontSize"] == 32) else "0xFFFFFFFFFFFFFFFF"
                                kStr += inst("s_mov_b{}".format(kernel["WavefrontSize"]), sgpr(tmpSgpr, writer.laneSGPRCount), all1mask, "to restore all threads active")
                                kStr += inst("s_or_saveexec_b{}".format(kernel["WavefrontSize"]), writer.vcc, sgpr(tmpSgpr,writer.laneSGPRCount), "all threads active")
                                kStr += inst("s_nop", "3", "wait for exec mask")

                    kStr += inst("s_branch label_%04u" % shiftLabels[glvw-1], "done" )

        kStr += "label_%04u: // end shift0%s" % (shiftLabels[glvw-1], writer.endLine)

        writer.vgprPool.checkIn(movReg)
        writer.vgprPool.checkIn(threadIdInCoalReg)
        writer.vgprPool.checkIn(permuteIndexReg)

        writer.vgprPool.checkIn(glvwblkidReg)
        writer.vgprPool.checkIn(sReg)
        writer.vgprPool.checkIn(wReg)
        writer.vgprPool.checkIn(mtReg)

        # checkin scratch vgprs
        writer.vgprPool.checkIn(tmpVgpr)
        writer.vgprPool.checkIn(wgMT)

        return kStr



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest

import Tensile.Component  as Component
import Tensile.Components as Components
from .test_Component import MockWriter

@pytest.fixture
def aggressive():
    return {"kernel": {'AggressivePerfMode': True}}

@pytest.fixture
def non_aggressive():
    return {'kernel': {'AggressivePerfMode': False}}

def test_aggressive(aggressive):
    writer = MockWriter(**aggressive)

    found = Component.Component.Priority.find(writer)
    assert isinstance(found, Components.Priority.AggressivePriority)

    firstRaise  = found(writer, 1, "comment")
    secondRaise = found(writer, 1, "comment")
    lower = found(writer, 0)

    assert "s_setprio 1" in firstRaise
    assert "comment" in firstRaise
    assert secondRaise == ""
    assert "s_setprio 0" in lower

def test_non_aggressive(non_aggressive):
    writer = MockWriter(**non_aggressive)

    found = Component.Component.Priority.find(writer)
    assert isinstance(found, Components.Priority.ConstantPriority)

    firstRaise  = found(writer, 1, "comment")
    secondRaise = found(writer, 1, "comment")
    lower = found(writer, 0)

    assert firstRaise == ""
    assert secondRaise == ""
    assert lower == ""



################################################################################
#
# Copyright (C) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import collections
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import List, Union

from .. import Utils
from ..Common import ensurePath, gfxName, globalParameters, printWarning, tPrint
from ..KernelWriterAssembly import KernelWriterAssembly
from ..SolutionStructs import Solution
from .SharedCommands import compressCodeObject


def _linkIntoCodeObject(
    objFiles: List[str],
    coPathDest: Union[Path, str],
    writer: KernelWriterAssembly,
    maxLineLength: int,
):
    """Links object files into a code object file.

    Args:
        objectFiles: A list of object files to be linked.
        coPathDest: The destination path for the code object file.
        writer: An instance of KernelWriterAssembly to get link arguments.
        maxLineLength: The maximum command line length. On Windows, this is nominally 8191,
            on posix platforms this number can be computed with ``$ getconf ARG_MAX``.

    Raises:
        RuntimeError: If linker invocation fails.
    """
    args = writer.getLinkCodeObjectArgs(objFiles, str(coPathDest))
    # It is possible for the list of `objFiles` to exceed the command line limit
    # LLVM allows the provision of compiler arguments via a "response file" (`rf` below)
    # Reference: https://llvm.org/docs/CommandLine.html#response-files
    lineLength = sum(len(arg) for arg in args) + len(args) - 1  # Account for spaces
    if lineLength > maxLineLength:
        with tempfile.NamedTemporaryFile(mode="wt", delete=False) as rf:
            # Use repr on Windows to a get raw string with non-escaped path separators
            strArgs = " ".join(map(repr, objFiles)) if os.name == "nt" else " ".join(objFiles)
            rf.write(strArgs)
            rf.flush()
            args = writer.getLinkCodeObjectArgs([f"@{rf.name}"], str(coPathDest))

    tPrint(2, "Linking objects into co files: " + " ".join(args))
    try:
        out = subprocess.check_output(args, stderr=subprocess.STDOUT)
        tPrint(2, f"Output: {out}")
    except subprocess.CalledProcessError as err:
        raise RuntimeError(
            f"Error linking object files in to code object: {err.output}\nFailed command: {' '.join(args)}"
        )


def buildAssemblyCodeObjectFiles(
    bundler: str,
    kernels: List[Solution],
    writer: KernelWriterAssembly,
    outputPath: Union[Path, str],
    removeTemporaries: bool,
):

    countAsmKernels = lambda kernels: sum(k["KernelLanguage"] == "Assembly" for k in kernels)

    extObj = ".o"
    extCo = ".co"
    extCoRaw = ".co.raw"

    destDir = Path(ensurePath(os.path.join(outputPath, "library")))
    asmDir = Path(writer.getAssemblyDirectory())

    maxLineLength = (
        8191 if os.name == "nt" else int(subprocess.check_output(["getconf", "ARG_MAX"]).strip())
    )

    assemblyKernels = [k for k in kernels if k["KernelLanguage"] == "Assembly"]
    if len(assemblyKernels) == 0:
        return []

    archKernelMap = collections.defaultdict(list)
    for k in assemblyKernels:
        archKernelMap[tuple(k["ISA"])].append(k)

    coFiles = []
    for arch, archKernels in archKernelMap.items():
        gfx = gfxName(arch)
        objectFiles = [
            str(asmDir / (writer.getKernelFileBase(k) + extObj))
            for k in archKernels
            if "codeObjectFile" not in k
        ]

        if countAsmKernels(archKernels) == 0:
            continue

        if (
            globalParameters["MergeFiles"]
            or globalParameters["NumMergedFiles"] > 1
            or globalParameters["LazyLibraryLoading"]
        ):

            # Group kernels from placeholder libraries
            coFileMap = collections.defaultdict(list)

            if len(objectFiles):
                coFileMap[asmDir / ("TensileLibrary_" + gfx + extCoRaw)] = objectFiles

            for kernel in archKernels:
                coName = kernel.get("codeObjectFile", None)
                if coName:
                    coFileMap[asmDir / (coName + extCoRaw)].append(
                        str(asmDir / (writer.getKernelFileBase(kernel) + extObj))
                    )

            for coFileRaw, objFiles in coFileMap.items():

                _linkIntoCodeObject(objFiles, coFileRaw, writer, maxLineLength)
                coFile = destDir / coFileRaw.name.replace(extCoRaw, extCo)
                compressCodeObject(coFileRaw, coFile, gfx, bundler)
                coFiles.append(coFile)

        else:
            # no mergefiles
            assemblyKernelNames = [writer.getKernelFileBase(k) for k in archKernels]
            origCOFiles = [os.path.join(asmDir, k + ".co") for k in assemblyKernelNames]
            newCOFiles = [os.path.join(destDir, k + "_" + gfx + ".co") for k in assemblyKernelNames]

            for src, dst in (
                zip(origCOFiles, newCOFiles)
                if globalParameters["PrintLevel"] == 0
                else Utils.tqdm(zip(origCOFiles, newCOFiles), desc="Relocating code objects")
            ):
                shutil.copyfile(src, dst)
            coFiles += newCOFiles
            printWarning("Code object files are not compressed in `--no-merge-files` build mode.")

    return coFiles



################################################################################
#
# Copyright (C) 2021-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import ComputeStoreVgprs
from ..AsmUtils import log2, vectorStaticDivideAndRemainder, staticMultiply, vgpr, sgpr, inst, vectorStaticDivide, vectorStaticRemainder

class ComputeStoreVgprsVALU(ComputeStoreVgprs):
    kernel = {"EnableMatrixInstructionStore": False}

    """
    computeStoreVgprs
    Compute workitem/TT offsets in VGPRS
    and coord0/coord1
    tid0Scale specifies the number of output elements in 0/coalesced dim
    that should be written by each work-item in each batch element.
    """
    def __call__(self, writer, kernel, divisor, tid0Scale, tid1Scale):

        kStr = ""

        tmpS0 = writer.getTmpSgpr(3).idx()
        tmpS1 = tmpS0+1
        wgMT1 = tmpS0+2

        if writer.prefetchAcrossPersistent:
            wg0="PrevWorkGroup0"
            wg1="PrevWorkGroup1"
        else:
            wg0="WorkGroup0"
            wg1="WorkGroup1"

        # tid0, tid1: element offsets from the start of macroTile in 0 and 1 direction
        # These will live for entire GlobalWrite loop - allocate before tmps
        # to avoid fragmentation
        tid0 = writer.vgprPool.checkOut(1, "tid0")
        tid1 = writer.vgprPool.checkOut(1, "tid1")

        packedC1 = kernel["PackedC1IndicesX"]

        if kernel["BufferStore"]:
            writer.cinRowPtr    = writer.vgprPool.checkOut(1, "cinRowPtr")
            writer.coutRowPtr = writer.vgprPool.checkOut(1, "coutRowPtr")

        kStr += vectorStaticDivideAndRemainder(tid1, tid0, "Serial", divisor, tmpS0)
        kStr += staticMultiply(vgpr(tid0), vgpr(tid0), tid0Scale, sgpr(tmpS1))
        if tid1Scale != 1:
            kStr += staticMultiply(vgpr(tid1), vgpr(tid1), tid1Scale, sgpr(tmpS1))

        if kernel["BufferStore"]:
            # compute rowStart- this is just tid1 scaled by appropriate stride.
            # rowPtr is offset from the beginning of the tile/SRD not the tensor base
            # when incremented, it moves in units of (col) Stride to get to a new row
            # it is used for address computation, not element range detection.
            # rowPtr is in the element space and must be scaled by bpe if bytes are required.
            # Do this before code below which overwries the tid1:
            # TODO-packed
            # Eventually need to modify if supporting packed coord1, to start just assert if that case is detected
            #--
            strideC1 = "StrideC%s" % (writer.indexChars[packedC1[0]])
            kStr += inst("v_mul_lo_u32", vgpr(writer.cinRowPtr),
                         vgpr(tid1), sgpr(strideC1), \
                         "rowStart vgpr")
            strideD1 = "StrideD%s" % (writer.indexChars[packedC1[0]])
            kStr += inst("v_mul_lo_u32", vgpr(writer.coutRowPtr),
                         vgpr(tid1), sgpr(strideD1), \
                         "rowStart vgpr")
            kStr += "\n"

            #kStr += writer.assert_ne(sgpr("WorkGroup1"),1)

        # Compute coord0 and coord1
        # These are element offsets from the beginning of the tensor.
        # These are 'flattened' meaning they span packed tensor dims.
        # They need to be preserved so can use in comparisons against
        # product-of-packed sizes to determine OOB cases. (for Edge tiles only)
        kStr += inst("s_mul_i32", \
                sgpr(tmpS0), \
                hex(kernel["MacroTile0"]), \
                sgpr(wg0), \
                "%s = wg0*MT0"%sgpr(tmpS0))

        # coord = tid*VW + workgroup offset
        kStr += inst("_v_add_co_u32", \
                vgpr(tid0), \
                writer.vcc, \
                sgpr(tmpS0), \
                vgpr(tid0), \
                "coord0 = tid0*VW + wg0*MT0")
        kStr += inst("s_mul_i32", \
                sgpr(wgMT1), \
                hex(kernel["MacroTile1"]), \
                sgpr(wg1), \
                "<- wg1*MT1")
        kStr += inst("_v_add_co_u32", \
                vgpr(tid1), \
                writer.vcc, \
                sgpr(wgMT1), \
                vgpr(tid1), \
                "coord1 = tid1*VW + wg1*MT1")

        if len(packedC1) > 1:
          kStr += writer.extractPackedCoord1ToRowStart(kernel, packedC1, tid1, 'D')

        writer.coord0 = tid0
        writer.coord1 = tid1

        return kStr

class ComputeStoreVgprsMFMA(ComputeStoreVgprs):
    kernel = {"EnableMatrixInstructionStore": True,
              "SourceSwap": False}

    """
    computeStoreVgprs
    Compute workitem/TT offsets in VGPRS
    and coord0/coord1
    tid0Scale specifies the number of output elements in 0/coalesced dim
    that should be written by each work-item in each batch element.
    """
    def __call__(self, writer, kernel, divisor, tid0Scale, tid1Scale):

        # writer.coord0
        # writer.coord1
        # writer.cinRowPtr  : C buffer column offset
        # writer.coutRowPtr : D buffer column offset

        # alloc resources
        tid0 = writer.vgprPool.checkOut(1)
        tid1 = writer.vgprPool.checkOut(1)
        if kernel["BufferStore"]:
            writer.cinRowPtr    = writer.vgprPool.checkOut(1, "cinRowPtr")
            writer.coutRowPtr = writer.vgprPool.checkOut(1, "coutRowPtr")

        wave_id = writer.vgprPool.checkOut(1)

        tmpVgpr0 = writer.vgprPool.checkOut(1,"tmpVgpr0")
        tmpSgpr  = writer.getTmpSgpr(1).idx()

        # constant
        MIBShape0 = kernel["MatrixInstM"] * kernel["MatrixInstBM"]
        MIBShape1 = kernel["MatrixInstN"] * kernel["MatrixInstBN"]

        # matrixInstM = kernel["MatrixInstM"] * kernel["MatrixInstBM"] if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstM"]
        matrixInstN = kernel["MatrixInstN"] * kernel["MatrixInstBN"] if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstN"]

        kStr = ""

        # coord 1 : wave part
        kStr += vectorStaticDivide(wave_id, "Serial", writer.kernel["WavefrontSize"], tmpSgpr)
        kStr += vectorStaticDivide(tid1, wave_id, kernel["MIWaveGroup"][0], tmpSgpr)
        kStr += inst("v_mul_lo_u32", vgpr(tid1), hex(MIBShape1), vgpr(tid1), "wave coordination offset 1")

        # coord 1 : thread part
        kStr += vectorStaticRemainder(tmpVgpr0, "Serial", matrixInstN, tmpSgpr)
        kStr += inst("_v_add_lshl_u32", vgpr(tid1), vgpr(tmpVgpr0), vgpr(tid1), log2(kernel["VectorWidthB"]), "coordination 1 = vwb *(wave_id1 + tid1)")

        if kernel["BufferStore"]:
          # coord 1 : offset part
          packedC1 = kernel["PackedC1IndicesX"]
          strideC1 = "StrideC%s" % (writer.indexChars[packedC1[0]])
          strideD1 = "StrideD%s" % (writer.indexChars[packedC1[0]])
          kStr += inst("v_mul_lo_u32", vgpr(writer.cinRowPtr), vgpr(tid1), sgpr(strideC1), " offset 1")
          kStr += inst("v_mul_lo_u32", vgpr(writer.coutRowPtr), vgpr(tid1), sgpr(strideD1), " offset 1")

        # coord 0 : thread part
        kStr += vectorStaticRemainder(tid0, "Serial", writer.kernel["WavefrontSize"], tmpSgpr)
        kStr += vectorStaticDivide(tid0, tid0, matrixInstN, tmpSgpr)
        kStr += staticMultiply(vgpr(tid0), vgpr(tid0), kernel["MIOutputVectorWidth"], sgpr(tmpSgpr), "thread0 * continuous_output")
        if kernel["MatrixInstBM"] > 1 and kernel["MatrixInstN"] == 4 and (kernel["MatrixInstM"] > kernel["MIOutputVectorWidth"]):
          # conversion for MI4x4 + MIBM>1
          # tid0 = (tid0/MIBM) + (tid0%MIBM)*(MIM//MIOVW)
          kStr += vectorStaticDivideAndRemainder(tmpVgpr0, tid0, tid0, kernel["MatrixInstBM"], tmpSgpr) # using rReg=dReg (assuming MIBM is power of 2)
          kStr += staticMultiply(vgpr(tid0), vgpr(tid0), kernel["MatrixInstM"]//kernel["MIOutputVectorWidth"], sgpr(tmpSgpr), "(tid1%MIBM)*(MIM//MIOVW)")
          kStr += inst("_v_add_u32", vgpr(tid0), vgpr(tmpVgpr0), vgpr(tid0), "tid0 = (tid0/MIBM) + (tid0%MIBM)*MIM")
        # coord 0 : wave part
        kStr += vectorStaticRemainder(tmpVgpr0, wave_id, kernel["MIWaveGroup"][0], tmpSgpr)
        kStr += inst("v_mul_lo_u32", vgpr(tmpVgpr0), hex(MIBShape0), vgpr(tmpVgpr0), "wave coordination offset 0")

        kStr += inst("_v_add_lshl_u32", vgpr(tid0), vgpr(tmpVgpr0), vgpr(tid0), log2(kernel["VectorWidthA"]), "coordination 0 = vwa *(wave_id0 + tid0)")

        if writer.prefetchAcrossPersistent:
            wg0="PrevWorkGroup0"
            wg1="PrevWorkGroup1"
        else:
            wg0="WorkGroup0"
            wg1="WorkGroup1"

        # macro tile 0 part
        kStr += inst("s_mul_i32", sgpr(tmpSgpr), kernel["MacroTile0"], sgpr(wg0), "wgp0 * MT0")
        kStr += inst("_v_add_u32", vgpr(tid0), sgpr(tmpSgpr), vgpr(tid0), "coord 0 = (tid0/MI_m)*4 + waveG0*MIB_m + MT0*SG0")

        # macro tile 1 part
        kStr += inst("s_mul_i32", sgpr(tmpSgpr), kernel["MacroTile1"], sgpr(wg1), "wgp1 * MT1")
        kStr += inst("_v_add_u32", vgpr(tid1), sgpr(tmpSgpr), vgpr(tid1), "coord 1 = (tid0%MI_m) + waveG1*MIB_n + MT1*SG1")

        # extract packed rowStart vgpr
        if kernel["BufferStore"] and len(packedC1) > 1:
          kStr += writer.extractPackedCoord1ToRowStart(kernel, packedC1, tid1, 'D')

        # release resource
        writer.vgprPool.checkIn(tmpVgpr0)
        writer.vgprPool.checkIn(wave_id)

        # StoreRemap: calculate
        # 1. local read address
        # 2. local write address
        # 3. global write coord0 and coord1
        if kernel["StoreRemapVectorWidth"]:
            kStr += writer.storeRemapComputeStoreVgprs(kernel)

        writer.coord0 = tid0
        writer.coord1 = tid1

        return kStr

class ComputeStoreVgprsMFMASwap(ComputeStoreVgprs):
    kernel = {"EnableMatrixInstructionStore": True,
              "SourceSwap": True}

    """
    computeStoreVgprs
    Compute workitem/TT offsets in VGPRS
    and coord0/coord1
    tid0Scale specifies the number of output elements in 0/coalesced dim
    that should be written by each work-item in each batch element.
    """
    def __call__(self, writer, kernel, divisor, tid0Scale, tid1Scale):

        # writer.coord0
        # writer.coord1
        # writer.cinRowPtr  : C buffer column offset
        # writer.coutRowPtr : D buffer column offset

        # alloc resources
        tid0 = writer.vgprPool.checkOut(1)
        tid1 = writer.vgprPool.checkOut(1)
        if kernel["BufferStore"]:
            writer.cinRowPtr    = writer.vgprPool.checkOut(1, "cinRowPtr")
            writer.coutRowPtr = writer.vgprPool.checkOut(1, "coutRowPtr")

        wave_id = writer.vgprPool.checkOut(1)

        tmpVgpr0 = writer.vgprPool.checkOut(1,"tmpVgpr0")
        tmpSgpr  = writer.getTmpSgpr(1).idx()

        # constant
        MIBShape0 = kernel["MatrixInstM"] * kernel["MatrixInstBM"]
        MIBShape1 = kernel["MatrixInstN"] * kernel["MatrixInstBN"]

        matrixInstM = kernel["MatrixInstM"] * kernel["MatrixInstBM"] if (kernel["MatrixInstM"] == 4) else kernel["MatrixInstM"]
        # matrixInstN = kernel["MatrixInstN"] * kernel["MatrixInstBN"] if (kernel["MatrixInstN"] == 4) else kernel["MatrixInstN"]

        kStr = ""

        kStr += vectorStaticDivide(wave_id, "Serial", writer.kernel["WavefrontSize"], tmpSgpr)


        # coord 1 : thread part
        kStr += vectorStaticRemainder(tid1, "Serial", writer.kernel["WavefrontSize"], tmpSgpr)
        kStr += vectorStaticDivide(tid1, tid1, matrixInstM, tmpSgpr)
        kStr += staticMultiply(vgpr(tid1), vgpr(tid1), kernel["MIOutputVectorWidth"], sgpr(tmpSgpr), "thread0 * continuous_output")
        if kernel["MatrixInstBN"] > 1 and kernel["MatrixInstM"] == 4 and (kernel["MatrixInstN"] > kernel["MIOutputVectorWidth"]):
          # conversion for MI4x4 + MIBN>1
          # tid1 = (tid1/MIBN) + (tid1%MIBN)*(MIN//MIOVW)
          kStr += vectorStaticDivideAndRemainder(tmpVgpr0, tid1, tid1, kernel["MatrixInstBN"], tmpSgpr) # using rReg=dReg (assuming MIBN is power of 2)
          kStr += staticMultiply(vgpr(tid1), vgpr(tid1), kernel["MatrixInstN"]//kernel["MIOutputVectorWidth"], sgpr(tmpSgpr), "(tid1%MIBN)*(MIN//MIOVW)")
          kStr += inst("_v_add_u32", vgpr(tid1), vgpr(tmpVgpr0), vgpr(tid1), "tid1 = (tid1/MIBN) + (tid1%MIBN)*MIN")
        # coord 1 : wave part
        kStr += vectorStaticDivide(tmpVgpr0, wave_id, kernel["MIWaveGroup"][0], tmpSgpr)
        kStr += inst("v_mul_lo_u32", vgpr(tmpVgpr0), hex(MIBShape1), vgpr(tmpVgpr0), "wave coordination offset 1")
        kStr += inst("_v_add_lshl_u32", vgpr(tid1), vgpr(tmpVgpr0), vgpr(tid1), log2(kernel["VectorWidthB"]), "coordination 1 = vwb *(wave_id1 + tid1)")

        if kernel["BufferStore"]:
          # coord 1 : offset part
          packedC1 = kernel["PackedC1IndicesX"]
          strideC1 = "StrideC%s" % (writer.indexChars[packedC1[0]])
          strideD1 = "StrideD%s" % (writer.indexChars[packedC1[0]])
          kStr += inst("v_mul_lo_u32", vgpr(writer.cinRowPtr), vgpr(tid1), sgpr(strideC1), " offset 1")
          kStr += inst("v_mul_lo_u32", vgpr(writer.coutRowPtr), vgpr(tid1), sgpr(strideD1), " offset 1")

        # coord 0 : wave part
        kStr += vectorStaticRemainder(tmpVgpr0, wave_id, kernel["MIWaveGroup"][0], tmpSgpr)
        if kernel["MIWaveGroup"][0] > 1:
          kStr += inst("v_mul_lo_u32", vgpr(tmpVgpr0), hex(MIBShape0), vgpr(tmpVgpr0), "wave coordination offset 0")

        # coord 0 : thread part
        kStr += vectorStaticRemainder(tid0, "Serial", matrixInstM, tmpSgpr)
        kStr += inst("_v_add_lshl_u32", vgpr(tid0), vgpr(tmpVgpr0), vgpr(tid0), log2(kernel["VectorWidthA"]), "coordination 0 = vwa *(wave_id0 + tid0)")

        if writer.prefetchAcrossPersistent:
            wg0="PrevWorkGroup0"
            wg1="PrevWorkGroup1"
        else:
            wg0="WorkGroup0"
            wg1="WorkGroup1"

        # macro tile 0 part
        kStr += inst("s_mul_i32", sgpr(tmpSgpr), kernel["MacroTile0"], sgpr(wg0), "wgp0 * MT0")
        kStr += inst("_v_add_u32", vgpr(tid0), sgpr(tmpSgpr), vgpr(tid0), "coord 0 = (tid0/MI_m)*4 + waveG0*MIB_m + MT0*SG0")

        # macro tile 1 part
        kStr += inst("s_mul_i32", sgpr(tmpSgpr), kernel["MacroTile1"], sgpr(wg1), "wgp1 * MT1")
        kStr += inst("_v_add_u32", vgpr(tid1), sgpr(tmpSgpr), vgpr(tid1), "coord 1 = (tid0%MI_m) + waveG1*MIB_n + MT1*SG1")

        # release resource
        writer.vgprPool.checkIn(tmpVgpr0)
        writer.vgprPool.checkIn(wave_id)

        # StoreRemap: calculate
        # 1. local read address
        # 2. local write address
        # 3. global write coord0 and coord1
        if kernel["StoreRemapVectorWidth"]:
            kStr += writer.storeRemapComputeStoreVgprs(kernel)

        writer.coord0 = tid0
        writer.coord1 = tid1

        return kStr



################################################################################
# Copyright 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell cop-
# ies of the Software, and to permit persons to whom the Software is furnished
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IM-
# PLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNE-
# CTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
################################################################################

from ..Component import MFMA
from ..DataType import DataType

class WMMASelection(MFMA):
    asmCaps = {"HasWMMA": True}

    def __call__(self, writer, accOutStart, accOutEnd, in0, in1, accInStart, accInEnd, accStoreCIdx, firstIter):
        kernel = writer.kernel
        inType = kernel["ProblemType"]["DataType"].toNameAbbrev()
        neg = " neg_lo:[1,1,1]" if (inType == "i8") else ""
        inType = "iu8" if inType == "i8" else inType
        outType = kernel["ProblemType"]["ComputeDataType"].toNameAbbrev()
        if kernel["ProblemType"]["DataType"].isComplex():
            inType = outType
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        miK = kernel["MatrixInstK"]
        # miB = kernel["MatrixInstB"]
        str0 = in1 if kernel["SourceSwap"] else in0
        str1 = in0 if kernel["SourceSwap"] else in1
        # use const 0 for src2 in firstIter case
        src2 = "0" if firstIter else "v[%u:%u]"%(accOutStart, accOutEnd)

        kStr = "v_wmma_%s_%ux%ux%u_%s v[%u+%u:%u+%u], %s, %s, %s%s%s" \
            % (outType, miM, miN, miK, inType, accInStart, accStoreCIdx, accInEnd, accStoreCIdx, str0, str1, src2, neg, writer.endLine)

        return kStr

class MFMASelection950(MFMA):
    versions = [(9,5,0)]

    def WaitCount(self, writer):
        kernel = writer.kernel
        dataType = kernel["ProblemType"]["DataType"]
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        if dataType.isSingle() or dataType.isSingleComplex() or dataType.isHalf() or dataType.isBFloat16():
            if miM == 4 and miN == 4:
                return 2
        elif dataType.isDouble() or dataType.isDoubleComplex():
            if miM == 4 and miN == 4:
                return 4
        return 0

    def __call__(self, writer, accOutStart, accOutEnd, in0, in1, accInStart, accInEnd, accStoreCIdx, firstIter):
        kernel = writer.kernel
        inType = kernel["ProblemType"]["DataType"].toNameAbbrev()
        # for F8 hybrid cases, we need to change the inType of VMFMA inst as well
        if kernel["SourceSwap"]:
            dataType = kernel["ProblemType"]["DataType"]
            if dataType.isFloat8BFloat8():
                inType = DataType("B8F8").toNameAbbrev() # change the intype from F8B8 to B8F8
            if dataType.isBFloat8Float8():
                inType = DataType("F8B8").toNameAbbrev() # change the intype from B8F8 to F8B8

        outType = kernel["ProblemType"]["DataType"].MIOutputTypeNameAbbrev()
        if kernel["ProblemType"]["DataType"].isComplex():
            inType = outType
        accType = "a" if not kernel["MIArchVgpr"] else "v"
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        miK = kernel["MatrixInstK"]
        miB = kernel["MatrixInstB"]
        str0 = in1 if kernel["SourceSwap"] else in0
        str1 = in0 if kernel["SourceSwap"] else in1

        strB = ""
        if miB > 1:
            strB = "%ub_" % miB

        # use const 0 for src2 in firstIter case
        src2 = "0" if firstIter else "%s[%u:%u]"%(accType, accOutStart, accOutEnd)

        kStr = "v_mfma_%s_%ux%ux%u_%s%s %s[%u+%u:%u+%u], %s, %s, %s%s" \
            % (outType, miM, miN, miK, strB, inType, accType, accInStart, accStoreCIdx, accInEnd, accStoreCIdx, str0, str1, src2, writer.endLine)

        return kStr

class MFMASelection940(MFMA):
    versions = [(9,4,2)]

    def WaitCount(self, writer):
        kernel = writer.kernel
        dataType = kernel["ProblemType"]["DataType"]
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        if dataType.isSingle() or dataType.isSingleComplex() or dataType.isHalf() or dataType.isBFloat16():
            if miM == 4 and miN == 4:
                return 2
        elif dataType.isDouble() or dataType.isDoubleComplex():
            if miM == 4 and miN == 4:
                return 4
        return 0

    def __call__(self, writer, accOutStart, accOutEnd, in0, in1, accInStart, accInEnd, accStoreCIdx, firstIter):
        kernel = writer.kernel
        inType = kernel["ProblemType"]["F32XdlMathOp"].toNameAbbrev() if kernel["EnableF32XdlMathOp"] else kernel["ProblemType"]["DataType"].toNameAbbrev()
        # for F8 hybrid cases, we need to change the inType of VMFMA inst as well
        if kernel["SourceSwap"]:
            dataType = kernel["ProblemType"]["DataType"]
            if dataType.isFloat8BFloat8():
                inType = DataType("B8F8").toNameAbbrev() # change the intype from F8B8 to B8F8
            if dataType.isBFloat8Float8():
                inType = DataType("F8B8").toNameAbbrev() # change the intype from B8F8 to F8B8

        outType = kernel["ProblemType"]["F32XdlMathOp"].MIOutputTypeNameAbbrev() if kernel["EnableF32XdlMathOp"] else kernel["ProblemType"]["DataType"].MIOutputTypeNameAbbrev()
        if kernel["ProblemType"]["DataType"].isComplex():
            inType = outType
        accType = "a" if not kernel["MIArchVgpr"] else "v"
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        miK = kernel["MatrixInstK"]
        miB = kernel["MatrixInstB"]
        str0 = in1 if kernel["SourceSwap"] else in0
        str1 = in0 if kernel["SourceSwap"] else in1

        strB = ""
        if miB > 1:
            strB = "%ub_" % miB

        # use const 0 for src2 in firstIter case
        src2 = "0" if firstIter else "%s[%u:%u]"%(accType, accOutStart, accOutEnd)

        kStr = "v_mfma_%s_%ux%ux%u_%s%s %s[%u+%u:%u+%u], %s, %s, %s%s" \
            % (outType, miM, miN, miK, strB, inType, accType, accInStart, accStoreCIdx, accInEnd, accStoreCIdx, str0, str1, src2, writer.endLine)

        return kStr

class MFMASelection(MFMA):
    versions = [(9,0,8), (9,0,10)]

    def __call__(self, writer, accOutStart, accOutEnd, in0, in1, accInStart, accInEnd, accStoreCIdx, firstIter):
        kernel = writer.kernel
        inType = "bf16" if kernel["ProblemType"]["Fp16AltImpl"] else kernel["ProblemType"]["DataType"].toNameAbbrev()
        outType = kernel["ProblemType"]["DataType"].MIOutputTypeNameAbbrev()
        if kernel["ProblemType"]["DataType"].isComplex():
            inType = outType
        accType = "a" if not kernel["MIArchVgpr"] else "v"
        mfma1k = "_1k" if (kernel["MFMA_BF16_1K"] or kernel["ProblemType"]["Fp16AltImpl"]) else ""
        miM = kernel["MatrixInstM"]
        miN = kernel["MatrixInstN"]
        miK = kernel["MatrixInstK"]
        str0 = in1 if kernel["SourceSwap"] else in0
        str1 = in0 if kernel["SourceSwap"] else in1

        # use const 0 for src2 in firstIter case
        src2 = "0" if firstIter else "%s[%u:%u]"%(accType, accOutStart, accOutEnd)

        kStr = "v_mfma_%s_%ux%ux%u%s%s %s[%u+%u:%u+%u], %s, %s, %s%s" \
                % (outType, miM, miN, miK, inType, mfma1k, accType, accInStart, accStoreCIdx, accInEnd, accStoreCIdx, str0, str1, src2, writer.endLine)

        return kStr



################################################################################
#
# Copyright (C) 2021-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from copy import deepcopy
import ast

class ReadWriteTransformDict(dict):  # dicts take a mapping or iterable as their optional first argument
    """
    The focus of this class is to allow explicit customization of
    reading and writing to attributes.
    Users of this class might restrict read or write access to particular
    variables, or perform some customization on those functionalities.
    """

    __slots__ = () # no __dict__ - that would be redundant

    readTransformFuncKey = '_readTransformFunc'
    writeTransformFuncKey = '_writeTransformFunc'

    def __init__(self, readTransformFunc = None, writeTransformFunc = None):
        super().__init__()
        if readTransformFunc is not None and callable(readTransformFunc):
            self.setReadTransform(readTransformFunc)
        if writeTransformFunc is not None and callable(writeTransformFunc):
            self.setWriteTransform(writeTransformFunc)

    def __getitem__(self, key):
        return self.readWithTransform(key) if self.hasReadTransform() else self.readNoTransform(key)

    def __setitem__(self, key, value):
        self.writeWithTransform(key, value) if self.hasWriteTransform() else self.writeNoTransform(key, value)

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def __copy__(self): # don't delegate w/ super - dict.copy() -> dict :(
        cls = self.__class__
        result = cls.__new__(cls)
        result.update(self)
        return result

    def __deepcopy__(self, memo): # don't delegate w/ super - dict.copy() -> dict :(
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for key in self.keys():
            result.writeNoTransform(key, deepcopy(self.readNoTransform(key)))
        return result

    def __repr__(self):
        return "\n".join(self.__toPrettyLines())

    def __toPrettyLines(self):
        """
        Embedded values are sometimes hard to read and understand
        with so many layers, so add some tabs and line breaks
        to help distinguish hierarchy trees of objects in the dict.
        """
        prettyLines = []
        prettyLines += ["<{0}(".format(type(self).__name__)]
        for key, value in self.items():
            if isinstance(value, ReadWriteTransformDict):
                valuePrettyLines = value.__toPrettyLines()
                prettyLines += ["\t{0}: {1}".format(key, valuePrettyLines.pop(0))]
                for line in valuePrettyLines:
                    prettyLines += ["\t{0}".format(line)]
            else:
                prettyLines += ["\t{0}: {1}".format(key, value)]
        prettyLines += [")>"]
        return prettyLines

    def get(self, key, default=None):
        try:
            return self[key]
        except:
            return default

    def set(self, key, value):
        try:
            self[key] = value
        except:
            pass

    # Transform tests
    def hasReadTransform(self):
        return self.readTransformFuncKey in self

    def hasWriteTransform(self):
        return self.writeTransformFuncKey in self

    # Omit transform
    def readNoTransform(self, name):
        return super().__getitem__(name)

    def writeNoTransform(self, name, value):
        super().__setitem__(name, value)

    # Use transform
    def readWithTransform(self, name):
        return self.getReadTransform()(self, name)

    def writeWithTransform(self, name, value):
        self.getWriteTransform()(self, name, value)

    # Transform get/set
    def getReadTransform(self):
        return self.readNoTransform(self.readTransformFuncKey) if self.hasReadTransform() else None

    def setReadTransform(self, readTransformFunc):
        self.writeNoTransform(self.readTransformFuncKey, readTransformFunc)
        if readTransformFunc is None:
            self.pop(self.readTransformFuncKey, None)

    def getWriteTransform(self):
        return self.readNoTransform(self.writeTransformFuncKey) if self.hasWriteTransform() else None

    def setWriteTransform(self, writeTransformFunc):
        self.writeNoTransform(self.writeTransformFuncKey, writeTransformFunc)
        if writeTransformFunc is None:
            self.pop(self.writeTransformFuncKey, None)

    def toDict(self):
        return dict(zip(self.keys(), self.values()))

    @staticmethod
    def flattenDict(obj, prefix = "", separator="."):
        """
        Sometimes it is favorable to reduce to a single hierarchy
        dict for readability or debugging.
        """
        result = ReadWriteTransformDict()
        for key in obj.keys():
            attribute = obj.readNoTransform(key)
            attribute = attribute.readNoTransform("value") if isinstance(attribute, ReadWriteTransformDict) else attribute
            if isinstance(attribute, dict):
                subDict = ReadWriteTransformDict.toFlattenedDict(attribute, key, separator)
                for subKey in subDict.keys():
                    newKey = separator.join([prefix, subKey]) if prefix != "" else subKey
                    result[newKey] = deepcopy(subDict[subKey])
            else:
                newKey = separator.join([prefix, key]) if prefix != "" else key
                result[newKey] = deepcopy(obj[key])
        return result

    def toFlattenedDict(self, prefix = "", separator="."):
        return ReadWriteTransformDict.flattenDict(self, prefix, separator)

class Parameter(ReadWriteTransformDict):
    """
    The Parameter class will be used as the main
    container for configuration properties.
    For this class we want to group configurations
    by name, value, default value and descriptions.
    No other attributes may be set.
    All operators  and reverse operators are overloaded
    to use value attribute as the operand as this is
    usually the one we are interested in.
    """
    __slots__ = ()

    # Simplification of creating attributes
    def createAttr(self, name, value):
        super().writeNoTransform(name, value)

    def __init__(self, name, initialValue, defaultValue=None, description=""):
        self.createAttr("type", type(initialValue))
        self.createAttr("name", name)
        self.createAttr("value", initialValue)
        self.createAttr("defaultValue", defaultValue)
        self.createAttr("description", description)

#        if defaultValue is None:
#            self.createAttr("defaultValue", initialValue)
#        else:
#            self.createAttr("defaultValue", defaultValue)

        # Restrict writing to only known fields
        def writeXForm(self, key, value):
            if key not in self or key == "type":
                raise AttributeError("Cannot write attribute: {0}".format(key))
            else:
                # Ensure incoming values are same type to allow assignment
                objType = self.readNoTransform("type")
                incomingType = type(value)
                if key == "value" and objType is not incomingType:
                    raise AttributeError("Type preservation: stored {0} != incoming {1}".format(objType, incomingType))
                else:
                    self.writeNoTransform(key, value)

        super().__init__(None, writeXForm)

    # Boolean ops
    def __lt__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value < rhs.value
        else:
            return self.value < rhs

    def __rlt__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value < self.value
        else:
            return lhs < self.value

    def __le__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value <= rhs.value
        else:
            return self.value <= rhs

    def __rle__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value <= self.value
        else:
            return lhs <= self.value

    def __eq__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value == rhs.value
        else:
            return self.value == rhs

    def __req__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value == self.value
        else:
            return lhs == self.value

    def __ne__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value != rhs.value
        else:
            return self.value != rhs

    def __rne__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value != self.value
        else:
            return lhs != self.value

    def __gt__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value > rhs.value
        else:
            return self.value > rhs

    def __rgt__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value > self.value
        else:
            return lhs > self.value

    def __ge__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value >= rhs.value
        else:
            return self.value >= rhs

    def __rge__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value >= self.value
        else:
            return lhs >= self.value

    # Binary Ops with their reverses
    def __add__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value + rhs.value
        else:
            return self.value + rhs

    def __radd__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value + self.value
        else:
            return lhs + self.value

    def __sub__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value - rhs.value
        else:
            return self.value - rhs

    def __rsub__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value - self.value
        else:
            return lhs - self.value

    def __mul__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value * rhs.value
        else:
            return self.value * rhs

    def __rmul__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value * self.value
        else:
            return lhs * self.value

    def __truediv__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value / rhs.value
        else:
            return self.value / rhs

    def __rtruediv__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value / self.value
        else:
            return lhs / self.value

    def __floordiv__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value // rhs.value
        else:
            return self.value // rhs

    def __rfloordiv__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value // self.value
        else:
            return lhs // self.value

    def __mod__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value % rhs.value
        else:
            return self.value % rhs

    def __rmod__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value % self.value
        else:
            return lhs % self.value

    def __pow__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value ** rhs.value
        else:
            return self.value ** rhs

    def __rpow__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value ** self.value
        else:
            return lhs ** self.value

    def __rshift__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value >> rhs.value
        else:
            return self.value >> rhs

    def __rrshift__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value >> self.value
        else:
            return lhs >> self.value

    def __lshift__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value << rhs.value
        else:
            return self.value << rhs

    def __rlshift__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value << self.value
        else:
            return lhs << self.value

    def __and__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value & rhs.value
        else:
            return self.value & rhs

    def __rand__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value & self.value
        else:
            return lhs & self.value

    def __or__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value | rhs.value
        else:
            return self.value | rhs

    def __ror__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value | self.value
        else:
            return lhs | self.value

    def __xor__(self, rhs):
        if isinstance(rhs, Parameter):
            return self.value ^ rhs.value
        else:
            return self.value ^ rhs

    def __rxor__(self, lhs):
        if isinstance(lhs, Parameter):
            return lhs.value ^ self.value
        else:
            return lhs ^ self.value

    def __bool__(self):
        return bool(self.value)

    # Unary ops
    def __neg__(self):
        return -self.value

    def __pos__(self):
        return +self.value

    def __invert__(self):
        return ~self.value

    # Interface
    def resetToDefault(self):
        self.value = self.defaultValue
        return self.value

    def getValue(self):
        return self.value

    def getDefault(self):
        return self.defaultValue

    def getDescription(self):
        return self.description

class CallableParameter(Parameter):
    """
    This class is a Parameter who gets its value
    dynamically at read time. It stores a function
    that gets called with self as context, which the
    result will replace the value attribute.
    This is useful for retrieving values that are not
    constants and must be retrieved from another function
    calls
    """
    __slots__ = ()

    class BadFunc(Exception):
        pass

    def __init__(self, name, callFunc, description=""):
        super().__init__(name, 0, 0, description)
        self.createAttr("callFunc", callFunc)

        # Before each read, call on self to update value
        rXForm = self.getReadTransform()
        def readXForm(obj, key):
            if key == "value":
                obj.writeNoTransform(key, obj.__call__())
            return rXForm(obj, key) if rXForm is not None else obj.readNoTransform(key)
        self.setReadTransform(readXForm)

        # Prevent external writing to value attribute
        # Will be updated by reads as callable is updated.
        wXForm = self.getWriteTransform()
        def writeXForm(obj, key, value):
            if key == "value":
                pass
            else:
                wXForm(obj, key, value) if wXForm is not None else obj.writeNoTransform(key, value)
        self.setWriteTransform(writeXForm)

    def __call__(self):
        return self.callFunc(self)

    @classmethod
    def createBinaryOp(cls, lhs, rhs, op):
        """
        This function will build a binary operation
        function to call with two operands that are
        bound to the parameter
        """
        FuncMap = {
        "And"      : lambda lhs, rhs : lhs and rhs,
        "Or"       : lambda lhs, rhs : lhs or rhs,
        "Lt"       : lambda lhs, rhs : lhs <  rhs,
        "LtE"      : lambda lhs, rhs : lhs <= rhs,
        "Eq"       : lambda lhs, rhs : lhs == rhs,
        "NotEq"    : lambda lhs, rhs : lhs != rhs,
        "Gt"       : lambda lhs, rhs : lhs >  rhs,
        "GtE"      : lambda lhs, rhs : lhs >= rhs,
        "Mult"     : lambda lhs, rhs : lhs * rhs,
        "Pow"      : lambda lhs, rhs : lhs ** rhs,
        "Div"      : lambda lhs, rhs : lhs / rhs,
        "FloorDiv" : lambda lhs, rhs : lhs // rhs,
        "Mod"      : lambda lhs, rhs : lhs % rhs,
        "Add"      : lambda lhs, rhs : lhs + rhs,
        "Sub"      : lambda lhs, rhs : lhs - rhs,
        "BitAnd"   : lambda lhs, rhs : lhs & rhs,
        "BitOr"    : lambda lhs, rhs : lhs | rhs,
        "BitXor"   : lambda lhs, rhs : lhs ^ rhs,
        "LShift"   : lambda lhs, rhs : lhs << rhs,
        "RShift"   : lambda lhs, rhs : lhs >> rhs,
        "min"      : lambda lhs, rhs : min(lhs, rhs),
        "max"      : lambda lhs, rhs : max(lhs, rhs),
        }

        # Incoming op is a name
        if isinstance(op, str):
            assert op in FuncMap, "Missing operation in funcMap: {0}".format(op)
            name = op
            func = FuncMap[op]
        # Incoming op assumed as a function
        else:
            name = "CustomBinaryOp"
            func = op

        # Verify that the function is callable with 2 params
        try:
            func(lhs, rhs)
        except Exception as ex:
            raise CallableParameter.BadFunc from ex

        # Capture function and attached values
        opKey = next((key for key in FuncMap if FuncMap[key] == op), None)
        if opKey is not None and opKey in ["And", "Or", "Lt", "Le", "Eq", "NotEq", "Gt", "Ge"]:
            callFunc = lambda obj : bool(func(obj.lhs, obj.rhs))
        else:
            callFunc = lambda obj : func(obj.lhs, obj.rhs)

        binOp = cls(name, callFunc, "Binary operaton with two operands")
        binOp.createAttr("lhs", lhs)
        binOp.createAttr("rhs", rhs)

        return binOp

    @classmethod
    def createUnaryOp(cls, rhs, op):
        """
        This function will build a unary operation
        function to call with one operand that is
        bound to the parameter
        """

        FuncMap = {
        "Not"    : lambda val : not val,
        "Invert" : lambda val : ~val,
        "USub"   : lambda val : -val,
        "UAdd"   : lambda val : +val,
        "None"   : lambda val : val
        }

        # Incoming op is a name
        if isinstance(op, str):
            assert op in FuncMap, "Missing operation in funcMap: {0}".format(op)
            name = op
            func = FuncMap[op]
        # Incoming op assumed as a function
        else:
            name = "CustomUnaryOp"
            func = op

        # Verify that the function is callable with 1 param
        try:
            func(rhs)
        except Exception as ex:
            raise CallableParameter.BadFunc from ex

        # Capture function and attached values
        opKey = next((key for key in FuncMap if FuncMap[key] == op), None)
        if opKey is not None and opKey in ["Not"]:
            callFunc = lambda obj : bool(func(obj.rhs))
        else:
            callFunc = lambda obj : func(obj.rhs)

        unOp = cls(name, callFunc, "Unary operaton with one operand")
        unOp.createAttr("rhs", rhs)

        return unOp

class ExpressionEvaluator(object):
    """
    This class is a recursive visitor of an ast module tree.
    The ast tree is generated from a text expression (e.g. a > 5).
    Given a context object to resolve names to values, the expression
    evaluator will return a chain of callable parameters that can be
    invoked to get the result of the expression.
    Example:
        context = {a : 10}
        tree = ast.parse("a > 5", mode='exec') # ast library
        exprEval = ExpressionEvaluator.evaluate(tree, context)
        result = exprEval.__call__() # = True
    """

    __slots__ = ()

    def evaluate(self, node, namesContext):

        #print("Type: {0}".format(type(node).__name__))
        #print(node)

        nodeType = type(node).__name__

        # Root Level 0 node (required)
        if nodeType == "Module":
            # fields: ('body')
            assert len(node.body) == 1, "Expecting only one expression"
            return self.evaluate(node.body[0], namesContext)

        # Level 1 node (required)
        elif nodeType == "Expr":
            # fields: ('value')
            result = self.evaluate(node.value, namesContext)
            return result.value if hasattr(result, "value") else result

        # Binary operations nodes which take 2 operands
        elif nodeType == "BinOp":
            # fields: ('op', 'left', 'right')
            lhs = self.evaluate(node.left, namesContext)
            rhs = self.evaluate(node.right, namesContext)
            op = type(node.op).__name__
            return CallableParameter.createBinaryOp(lhs, rhs, op)

        elif nodeType == "BoolOp":
            # fields: ('op', 'values')
            lhs = self.evaluate(node.values[0], namesContext)
            rhs = self.evaluate(node.values[1], namesContext)
            op = type(node.op).__name__
            return CallableParameter.createBinaryOp(lhs, rhs, op)

        elif nodeType == "Compare":
            # fields: ('left', 'ops', 'comparators')
            lhs = self.evaluate(node.left, namesContext)
            for i in range(len(node.ops)):
                rhs = self.evaluate(node.comparators[i], namesContext)
                op = type(node.ops[i]).__name__
                result = CallableParameter.createBinaryOp(lhs, rhs, op)
                lhs = result
            return lhs

        # Unary operations which take 1 operand
        elif nodeType == "UnaryOp":
            # fields: ('op', 'operand')
            value = self.evaluate(node.operand, namesContext)
            op = type(node.op).__name__
            return CallableParameter.createUnaryOp(value, op)

        elif nodeType == "Call":
            # fields: ('func', 'args', 'keywords')
            if len(node.args) == 2:
                lhs = self.evaluate(node.args[0], namesContext)
                rhs = self.evaluate(node.args[1], namesContext)
                # For the op, give empty context so we get a string
                # from the name node. Then pass it into the factory
                op = self.evaluate(node.func, {})
                return CallableParameter.createBinaryOp(lhs, rhs, op)

            elif len(node.args) == 1:
                rhs = self.evaluate(node.args[0], namesContext)
                # For the op, give empty context so we get a string
                # from the name node. Then pass it into the factory
                op = self.evaluate(node.func, {})
                return CallableParameter.createUnaryOp(rhs, op)

            assert 0, "Unknown function call with {0} parameters".format(len(node.args))

        elif nodeType == "Assign":
            # fields: ('targets', 'value')

            value = self.evaluate(node.value, namesContext)
            valueToAssign = value.value if hasattr(value, "value") else value

            for target in node.targets:
                targetType = type(target).__name__

                # If we are assigning to a name, use
                # current context as write object.
                if targetType == "Name":
                    assignObj = namesContext
                    assignAttr = target.id

                # If we are assigning to an attribute,
                # get the embedded context to write to
                elif targetType == "Attribute":
                    assignObj = self.evaluate(target, namesContext)
                    assignAttr = target.attr

                else: assert 0, "Don't know how to handle target node type: {0}".format(targetType)

                # The assignment value might be an expression,
                # so if it is callable, then evaluate it
                assignObj[assignAttr] = valueToAssign

            return valueToAssign

        # Leaf node types
        elif nodeType == "Name":
            # fields:  ('id')
            if node.id in namesContext:
                return namesContext[node.id]
            else:
                 print("No context for named variable: {0}".format(node.id))
                 return str(node.id)

        elif nodeType == "Attribute":
            # fields:  ('value', 'attr', 'ctx')

            # node.attr has the leaf attribute. We need to recurse and get
            # the leaf attribute container so we can access the attribute.
            # node.value has the next chained node, which is either another Attribute
            # or Name node
            newContext = self.evaluate(node.value, namesContext)
            assert node.attr in newContext, "No attribute for named variable: {0}".format(node.attr)

            # For storing context, return the target value's container so we can
            # setattr on it. See: Assign case.
            # Otherwise return the actual stored value
            return newContext if type(node.ctx).__name__ == "Store" else newContext[node.attr]

        elif nodeType == "IfExp":
            # fields ('test', 'body', 'orelse')
            theTest = self.evaluate(node.test, namesContext)
            trueCond = self.evaluate(node.body, namesContext)
            falseCond = self.evaluate(node.orelse, namesContext)
            return trueCond if theTest() else falseCond

        elif nodeType == "Num":
            # fields ('n')
            return node.n

        elif nodeType == "Constant":
            # fields ('kind', 'value')
            return node.value

        elif nodeType == "Str":
            # fields ('s')
            return node.s

        assert 0, "Unknown node type: {0}".format(nodeType)

# Everything stored as a Parameter object, which has
# value, default value and type attributes
class ProjectConfig(ReadWriteTransformDict):
    """
    Adds support for hierarchical configuration designs.
    Allows grouping of configurations together.
    Also allows dynamic constraints.

    Each Config entry is a Parameter type object that
    encapsulates attributes of Name, Type, Value, DefaultValue
    and Description as seen above.

    Example:

    Program:
        Network:
            IP: X.X.X.X
            PORT: XXXX
            MaxPort: YYYY
        UI:
            WindowSize: 2000
            Antialiasing: True

    # Creation patterns:
    Program = ProjectConfig()
    networkConfig = Program.createSection("Network")
    networkConfig.createValue("IP", X.X.X.X, "This is the IP we connect to")
    networkConfig.createValue("PORT", XXXX, "Port we want to access")
    networkConfig.createValue("PORT", YYYY, "Maximum port value")

    uiConfig = Program.createSection("UI")
    uiConfig.createValue("WindowSize", 2000)
    uiConfig.createValue("Antialiasing", True)

    # Access patterns:
    IP = Program.Network.IP
    PORT = Program["Network"]["IP"]
    networkConfig = Program["Network"]

    # Constraints assertions: e.g. limit port numbers or antialiasing.
    Program.addConstraint("Program.Network.PORT >= MaxPort")
    Program.addConstraint("False if Program.UI.WindowSize > 2000 and Program.UI.Antialiasing else True")

    # Overrides: e.g. enforce port numbers
    Program.addConstraint("Program.Network.PORT = Program.Network.MaxPort if Program.Network.PORT > Program.Network.MaxPort")

    """

    __slots__ = ()

    ContainerType = Parameter
    SubConfigType : 'ProjectConfig'
    ConstraintsKey = "_Constraints"

    def __init__(self):

        # Sub-configs will be stored inside ContainerType,
        # so when reading them as an attribute, forward the config
        # stored in the container's value
        def readXForm(obj, key):

            attribute = obj.readNoTransform(key)

            # Attribute is a Container, which has a "value"
            # and should be forwarded to get around awkward
            # value forwarding. We are mainly interested
            # in accessing the value attribute, so we can
            # now access it with config["Attrib"] instead
            # of config["Attrib"]["value"]
            if isinstance(attribute, self.ContainerType):
                return attribute.readNoTransform("value")
            else:
                return attribute

        def writeXForm(obj, key, value):
            if key in super(obj.__class__, obj).keys() and isinstance(obj.readNoTransform(key), self.ContainerType):
                obj.readNoTransform(key).writeNoTransform("value", value)
            else:
                obj.writeNoTransform(key, value)

        super().__init__(readXForm, writeXForm)

    def __contains__(self, key):
        return True if super().__contains__(key) else self.toFlattenedDict().__contains__(key)

    def __getContainer(self, key):
        # In case we are accessing with keys such as [a.b.c]
        # Then we need to recurse to get the embedded container.
        # Assumes that a and b for example are embedded configs
        if "." in key:
            try:
                levels = key.split(".")
                currentValue = self
                for level in levels[:-1]:
                    currentValue = currentValue[level]
                return currentValue.readNoTransform(levels[-1])
            except:
               pass

        # Fallback, in case there is a "." in the key
        return self.readNoTransform(key)

    def __getitem__(self, key):

        if "." in key:
            try:
                levels = key.split(".")
                currentValue = self
                for level in levels:
                    currentValue = currentValue[level]
                return currentValue
            except:
                pass

        # Fallback is to check if the full key lives
        # in the top level, in case it just happens
        # to have a '.' in it.
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        # In case we are accessing with keys such as [a.b.c]
        # Then we need to recurse to get the embedded value.
        # Assumes that a and b for example are embedded configs
        if "." in key:
            try:
                levels = key.split(".")
                currentValue = self
                for level in levels[:-1]:
                    currentValue = currentValue[level]
                setattr(currentValue, levels[-1], value)
                return
            except:
                pass

        # Fallback is to check if the full key lives
        # in the top level, in case it just happens
        # to have a '.' in it.
        super().__setitem__(key, value)

    # Simplification of creating attributes
    def createValue(self, name, value, defaultValue = None, description=""):
        self[name] = self.ContainerType(name, value, defaultValue, description)
        return self[name]

    def createSection(self, name):
        return self.createValue(name, ProjectConfig())

    def resetToDefaults(self):
        for key in self.keys():
            attribute = self.readNoTransform(key)
            if isinstance(attribute, ProjectConfig):
                attribute.resetToDefaults()
            elif isinstance(attribute, self.ContainerType):
                attribute.writeNoTransform("value", attribute.readNoTransform("defaultValue"))

    def addConstraint(self, expressionStr):
        if self.ConstraintsKey not in self:
            self.createSection(self.ConstraintsKey)

        self[self.ConstraintsKey].createValue(expressionStr, ast.parse(expressionStr, mode='exec'))

    def checkConstraints(self):
        result = True
        if self.ConstraintsKey in self:
            constraints = self.readNoTransform(self.ConstraintsKey).value
            for expression, tree in [(x, y) for (x, y) in constraints.items() if "_" not in x]:
                result = ExpressionEvaluator().evaluate(tree.value, self)
                value = result.value if isinstance(result, Parameter) else result
                assert value, "Constraint evaluation failed: {0}".format(expression)
        return value

    def getDefaultValue(self, name):
        return self.__getContainer(name).getDefault()

    def getDescription(self, name):
        return self.__getContainer(name).getDescription()



################################################################################
#
# Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import itertools
import os
from typing import Any, Callable

from .Utilities.ConditionalImports import joblib


def CPUThreadCount(enable=True):
    from .Common import globalParameters

    if not enable:
        return 1
    else:
        if os.name == "nt":
            # Windows supports at most 61 workers because the scheduler uses
            # WaitForMultipleObjects directly, which has the limit (the limit
            # is actually 64, but some handles are needed for accounting).
            cpu_count = min(os.cpu_count(), 61)
        else:
            cpu_count = len(os.sched_getaffinity(0))
        cpuThreads = globalParameters["CpuThreads"]
        if cpuThreads < 1:
            return min(cpu_count, 64)  # Max build threads to avoid out-of-memory
        return min(cpu_count, cpuThreads)


def OverwriteGlobalParameters(newGlobalParameters):
    from . import Common

    Common.globalParameters.clear()
    Common.globalParameters.update(newGlobalParameters)


def pcallWithGlobalParamsMultiArg(f, args, newGlobalParameters):
    OverwriteGlobalParameters(newGlobalParameters)
    return f(*args)


def pcallWithGlobalParamsSingleArg(f, arg, newGlobalParameters):
    OverwriteGlobalParameters(newGlobalParameters)
    return f(arg)


def ParallelMap(
    function: Callable,
    objects: Any,
    message: str = "",
    enable: bool = True,
    multiArg: bool = True,
) -> list:
    """Executes a function over a list of objects in parallel or sequentially.

    This function is generally equivalent to ``list(map(function, objects))``. However, it provides
    additional functionality to run in parallel, depending on the 'enable' flag and available CPU
    threads.

    Args:
        function: The function to apply to each item in 'objects'. If 'multiArg' is True, 'function'
                  should accept multiple arguments.
        objects: An iterable of objects to be processed by 'function'. If 'multiArg' is True, each
                 item in 'objects' should be an iterable of arguments for 'function'.
        message: Optional; a message describing the operation. Default is an empty string.
        enable: Optional; if False, disables parallel execution and runs sequentially. Default is True.
        multiArg: Optional; if True, treats each item in 'objects' as multiple arguments for
                  'function'. Default is True.
        verbose: Optional; verbosity level for parallel execution. Default is 0.

    Returns:
        A list containing the results of applying **function** to each item in **objects**.
    """

    from . import Utils
    from .Common import globalParameters

    threadCount = CPUThreadCount(enable)

    message += (
        f": {threadCount} thread(s)" + f", {len(objects)} tasks"
        if hasattr(objects, "__len__")
        else ""
    )

    if threadCount <= 1 or joblib is None:
        f = lambda x: function(*x) if multiArg else function(x)
        return [f(x) for x in Utils.tqdm(objects, desc=message)]

    inputs = list(zip(objects, itertools.repeat(globalParameters)))
    pargs = Utils.tqdm(inputs, desc=message)
    pcall = pcallWithGlobalParamsMultiArg if multiArg else pcallWithGlobalParamsSingleArg

    return joblib.Parallel(n_jobs=threadCount)(
        joblib.delayed(pcall)(function, a, params) for a, params in pargs
    )
